import json
import logging
import re
from json import JSONDecodeError
from operator import itemgetter
from typing import Type

from langchain_core.exceptions import OutputParserException
from langchain_core.messages import SystemMessage
from langchain_core.output_parsers import BaseOutputParser, StrOutputParser
from langchain_core.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
)
from langchain_core.runnables import RunnablePassthrough, RunnableLambda, RunnableBranch
from langchain_openai import ChatOpenAI
from pydantic import ValidationError

from voice_stream.langchain_extensions.memory import LastKMessagesPlaceholder
from voice_stream.langchain_extensions.runnables import RunnableLogger
from voice_stream.types import T

logger = logging.getLogger(__name__)


def reduce_schema(schema: dict) -> dict:
    if "title" in schema:
        del schema["title"]
    if "type" in schema:
        del schema["type"]
    for prop in schema["properties"]:
        if "title" in schema["properties"][prop]:
            del schema["properties"][prop]["title"]
    return schema


class PydanticV2OutputParser(BaseOutputParser):
    """Parse an output using a pydantic model."""

    pydantic_model: Type[T]
    """The pydantic model to parse."""

    def parse(self, text: str) -> T:
        # Greedy search for 1st json candidate.
        match = re.search(
            r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL
        )
        json_str = match.group() if match else ""
        try:
            json_object = json.loads(json_str, strict=False)
        except JSONDecodeError:
            # See if we can fix by replacing single quotes with double quotes.
            try:
                json_object = json.loads(json_str.replace("'", '"'), strict=False)
            except JSONDecodeError as e:
                raise ValueError(
                    f"Could not parse JSON: {json_str}, extracted from {text}", e
                )
        try:
            return self.pydantic_model.parse_obj(json_object)
        except ValidationError as e:
            errors = e.errors(
                include_url=False, include_context=True, include_input=True
            )

            def format_error(e):
                input = f"'{e['input']}' - " if e["msg"] != "Field required" else ""
                return f"{', '.join(e['loc'])}: {input}{e['msg']}"

            reformatted = "\n".join([format_error(e) for e in errors])
            raise OutputParserException(
                error=reformatted, observation=json_object, llm_output=text
            )

    def get_format_instructions(self) -> str:
        return PYDANTIC_FORMAT_INSTRUCTIONS.format(output_schema=self._output_schema())

    def _output_schema(self) -> str:
        schema = self.pydantic_model.schema()

        # Remove extraneous fields.
        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        # Ensure json in context is well-formed with double quotes.
        return json.dumps(reduced_schema)

    @property
    def _type(self) -> str:
        return "pydantic"

    def get_parsing_prompt(self):
        return ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content=PARSING_PROMPT.format(
                        model_name=self.pydantic_model.__name__,
                        output_schema=self._output_schema(),
                    )
                ),
                LastKMessagesPlaceholder(variable_name="history", max_messages=10),
                HumanMessagePromptTemplate.from_template(
                    "Existing Parse:\n{parse}\n\nErrors:\n{errors}\n\nUser message:\n{query}"
                ),
            ]
        )

    def get_next_question_prompt(self):
        return ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content=NEXT_QUESTION_PROMPT.format(
                        model_name=self.pydantic_model.__name__,
                        output_schema=self._output_schema(),
                    )
                ),
                LastKMessagesPlaceholder(variable_name="history", max_messages=20),
                HumanMessagePromptTemplate.from_template(
                    "{query}\n\nADDITIONAL INFO:\nParse:\n{parse}\n\nErrors:\n{errors}"
                ),
            ]
        )

    def get_confirmation_prompt(self):
        return ChatPromptTemplate.from_messages(
            [
                SystemMessage(
                    content=CONFIRMATION_PROMPT.format(
                        model_name=self.pydantic_model.__name__,
                        output_schema=self._output_schema(),
                    )
                ),
                HumanMessagePromptTemplate.from_template("{parsed_result}"),
            ]
        )

    def get_chain(self):
        next_question_chain = (
            {
                "query": itemgetter("query"),
                "history": itemgetter("history"),
                "parse": lambda x: x["parsed_result"]["exception"].observation,
                "errors": lambda x: x["parsed_result"]["exception"].args[0],
            }
            | self.get_next_question_prompt()
            # | RunnableLogger("Next question generator")
            | ChatOpenAI(model="gpt-4")
            | StrOutputParser()
        )
        confirmation_chain = (
            self.get_confirmation_prompt()
            # | RunnableLogger("Confirmation generator")
            | ChatOpenAI(model="gpt-4")
            | StrOutputParser()
        )

        def extract(x, field, default):
            # logger.info(f"Extracting {field} from {x}")
            return x.get(field, default)

        return (
            {
                "query": itemgetter("query"),
                "history": lambda x: x.get("history", []),
                "parse": lambda x: extract(x, "parse", "{}"),
                "errors": lambda x: extract(x, "errors", ""),
            }
            # Take the initial user message and try to parse it.  We either get a full object back, or an exception with the missing info
            | RunnablePassthrough.assign(
                parsed_result=(
                    self.get_parsing_prompt()
                    # | RunnableLogger("Parse input")
                    | ChatOpenAI(model="gpt-4")
                    | self
                ).with_fallbacks(
                    [RunnableLambda(lambda x: x)],
                    exceptions_to_handle=(OutputParserException,),
                    exception_key="exception",
                )
            )
            # If we didn't have a clean parse, then generate the next question.
            | RunnablePassthrough.assign(
                output=
                # RunnableLogger("Parse check") |
                RunnableBranch(
                    [
                        lambda x: "parsed_result" in x
                        and "exception" in x["parsed_result"],
                        next_question_chain,
                    ],
                    confirmation_chain,
                )
            )
        )


PYDANTIC_FORMAT_INSTRUCTIONS = """The output should be formatted as a JSON object that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{output_schema}
```"""


PARSING_PROMPT = """
You are an expert at extracting information from free text user responses into a structured JSON objects that conform to a specific 
schema.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

The schema you are using for this task is:   
```
{output_schema}
```

The object is a "{model_name}"

You may not have enough information from the user to fill in all fields.  That is totally fine.  You can produce a partial
object, which will later be filled in with other values.  Omit any fields where you aren't sure you have the right information.  Be conservative.

You will be given several inputs.  
- A "Previous Parse" which contains a partial version of the JSON object created from earlier in the conversation.
- A list of zero or more "Errors", which are problems with the "previous parse" object that prevented it from being correctly parsed. 
- A previous AI chat message, which is the question which was asked to the user based on the previous output.  It might be empty if this is the first parse.
- A "User message", which is what the user responded with.

Your job is to output a new JSON object that updates the "Previous Parse" based on the user message.
"""


NEXT_QUESTION_PROMPT = """
You are working in a system extracting information from free text user responses into a structured JSON objects that conform to a specific 
schema.  

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

The schema you are using for this task is:   
```
{output_schema}
```

The object is a "{model_name}"

Your job is to understand what information you need to successfully complete the structured JSON object, and then figure 
out the right next question to ask the user to elicit that information.  Only ask one question at a time.

You will be given the recent chat history for context.

The last user message in the context will have some extra information appended to it.
- A "Parse" which contains the current parsed version of the object based on the user's responses.
- A list of zero or more "Errors", which are problems with the "Parse" object.  This includes incorrect or missing fields. 

Your output should be the next question to ask the user.  You should only ask a one question, and that question should only ask for 
one piece of information.  If there are multiple errors, just try to fix one of them.  For example, if there are two fields missing, 
such as name and address, you could ask "What is your name?" or "What is your address?"  You should not ask for both.


"""

CONFIRMATION_PROMPT = """The previous conversation has taken information from the user and used it to construct a JSON object.

The object is a "{model_name}", and it's schema is:
```
{output_schema}
```

Your job is to take the object present in the message, and provide a confirmation response summarizing the information.  
The summary should be conversational and using normal prose.  Don't just list out all the fields and values.
"""



"""
Integration with Google Cloud APIs.
"""

import logging
from typing import AsyncIterator, Union, Tuple

import asyncstdlib
from google.api_core.exceptions import Cancelled
from google.cloud.speech_v1 import (
    SpeechAsyncClient as SpeechAsyncClientV1,
    StreamingRecognizeRequest as StreamingRecognizeRequestV1,
    StreamingRecognizeResponse as StreamingRecognizeResponseV1,
    StreamingRecognitionConfig as StreamingRecognitionConfigV1,
    RecognitionConfig as RecognitionConfigV1,
)
from google.cloud.speech_v2 import (
    StreamingRecognizeRequest,
    StreamingRecognitionConfig,
    StreamingRecognitionFeatures,
    SpeechAsyncClient,
    StreamingRecognizeResponse,
    RecognitionConfig,
    RecognitionFeatures,
)
from google.cloud.texttospeech_v1 import (
    TextToSpeechAsyncClient,
    SynthesizeSpeechRequest,
    SynthesisInput,
    VoiceSelectionParams,
    AudioConfig,
    AudioEncoding,
)
from pydantic import BaseModel

from voice_stream.audio import AudioFormat, remove_wav_header
from voice_stream.core import (
    map_step,
    partition_step,
    async_init_step,
)
from voice_stream.events import (
    SpeechStart,
    SpeechEnd,
    BaseEvent,
    SpeechPartialResult,
    SpeechToTextResult,
)
from voice_stream.integrations._google_endless import EndlessStream
from voice_stream.integrations.google_utils import (
    _resolve_audio_decoding,
    GoogleDecodingConfig,
)
from voice_stream.text_to_speech import AudioWithText

# Can be one of the standard audio formats or a Google AudioConfig object
GoogleAudioConfig = Union[AudioFormat, AudioConfig]

logger = logging.getLogger(__name__)


class TTSRequest(BaseModel):
    text: str
    voice: str


async def google_text_to_speech_step(
    async_iter: AsyncIterator[Union[str, TTSRequest]],
    text_to_speech_async_client: TextToSpeechAsyncClient,
    voice_name: str = "en-US-Standard-H",
    audio_format: GoogleAudioConfig = AudioFormat.OGG_OPUS,
    speaking_rate: float = 1.0,
) -> AsyncIterator[AudioWithText]:
    """
    Data flow step that converts text to speech using Google's Text-to-Speech service.

    This function takes in strings or TTSRequest objects and converts each into audio using the specified voice and
    audio format. The function supports customization of the voice for each item if provided within a TTSRequest object.

    Parameters
    ----------
    async_iter : AsyncIterator[Union[str, TTSRequest]]
        An asynchronous iterator over text blocks or TTSRequest objects.  Using TTSRequest objects allows the voice to be
        customized for each block of text.
    text_to_speech_async_client : TextToSpeechAsyncClient
        An instance of TextToSpeechAsyncClient for interacting with the Google Text-to-Speech API.
    voice_name : str, optional
        Default voice name to be used for text-to-speech conversion. Default is "en-US-Standard-H".
    audio_format : GoogleAudioConfig, optional
        The audio format for the output speech. Default is AudioFormat.OGG_OPUS.
    speaking_rate
         Speaking rate/speed, in the range [0.25, 4.0]. 1.0 is the normal native speed supported by the specific voice.
         2.0 is twice as fast, and 0.5 is half as fast. If unset(0.0), defaults to the native 1.0 speed.
         Ignored if a GoogleAudioConfig is passed to `audio_format`.

    Yields
    ------
    AsyncIterator[AudioWithText]
        An asynchronous iterator yielding AudioWithText objects, each containing
        the audio output and the original text.

    Notes
    -----
    - The function allows for dynamic voice selection if a TTSRequest object is provided,
      which specifies the voice for a particular text block.
    - Supports different audio encoding formats based on the GoogleAudioConfig.
    """
    audio_config = _resolve_google_audio_config(audio_format, speaking_rate)

    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if isinstance(item, TTSRequest):
                voice = item.voice
                text = item.text
            else:
                voice = voice_name
                text = item
            # noinspection PyTypeChecker
            request = SynthesizeSpeechRequest(
                input=SynthesisInput(text=text),
                voice=VoiceSelectionParams(
                    language_code="en-us",
                    name=voice,
                ),
                audio_config=audio_config,
            )
            result = await text_to_speech_async_client.synthesize_speech(request)
            audio = result.audio_content
            if audio_config.audio_encoding == AudioEncoding.MULAW:
                audio = remove_wav_header(audio)
            yield AudioWithText(audio=audio, text=text, audio_format=audio_format)


def google_speech_step(
    async_iter: AsyncIterator[bytes],
    speech_async_client: SpeechAsyncClient,
    project,
    location,
    recognizer,
    model: str = "latest_long",
    language_codes: Union[str, list[str]] = ["en-US", "es-US"],
    audio_format: Union[GoogleDecodingConfig, AudioFormat] = None,
    include_events: bool = False,
    max_minutes: int = 9,
    stream_reset_timeout_secs=240,
) -> Union[AsyncIterator[str], Tuple[AsyncIterator[str], AsyncIterator[BaseEvent]]]:
    """
    Data flow step for converting audio into text using Google Cloud Speech-to-Text V2 API.

    This function processes an asynchronous stream of audio bytes, using Google Cloud
    Speech-to-Text service to convert the audio into text. It supports additional
    configuration such as specifying the model, language codes, and audio format.
    If 'include_events' is set to True, it also returns a stream of speech recognition
    events alongside the text (:class:`~voice_stream.events.SpeechStart` and :class:`~voice_stream.events.SpeechEnd`).

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator over audio data in bytes.
    speech_async_client : SpeechAsyncClient
        An instance of SpeechAsyncClient for interacting with the Google Cloud Speech-to-Text API.
    project :
        The Google Cloud project identifier.
    location :
        The location or region of the Google Cloud project.
    recognizer :
        The recognizer identifier within the Google Cloud project.  This must be a previously created recognizer.  See
        :func:`~voice_stream.integrations.google_utils.create_recognizer` for details.
    model : str, optional
        The model to be used by the recognizer. Default is "latest_long".
    language_codes : Union[str, list[str]], optional
        The language code(s) for the recognizer. Can be a single string or a list of strings.
        Default is ["en-US", "es-US"].
    audio_format : GoogleDecodingConfig | AudioFormat, optional
        Optional configuration for audio decoding. Not required if the recognizer auto-detects the format.
    include_events : bool, optional
        If True, the function also returns a stream of speech recognition events. Default is False.
    max_minutes : int, optional
        The maximum number of minutes to process.  Default is 9.
    stream_reset_timeout_secs : int, optional
        The time internal on which to refresh the call to the recognizer.  Google only supports streams up to 5 minutes,
        so the default is 4 minutes.

    Returns
    -------
    Union[AsyncIterator[str], Tuple[AsyncIterator[str], AsyncIterator[voice_stream.events.BaseEvent]]]
        If `include_events` is False, returns an asynchronous iterator yielding recognized text from the audio stream.
        If `include-events` is True, returns a tuple with 2 iterators.  The first yields the recognized text, and the
        second contain speech events.

    Notes
    -----
    - The function breaks the audio stream into chunks, sends them to the Speech-to-Text
      API, and processes the responses to extract the transcript.
    - Speech recognition events include information like word timings and confidences.
    - If you intend to run a recognition stream that is longer than `stream_reset_timeout_secs`, you need to set the
      audio format.
    """
    initial_config = _initial_recognition_config(
        include_events,
        project,
        location,
        recognizer,
        model,
        language_codes,
        audio_format,
    )

    def map_audio(x):
        return StreamingRecognizeRequest(recognizer=recognizer, audio=x)

    stream = EndlessStream(
        speech_async_client,
        initial_config,
        audio_format=audio_format,
        map_audio=map_audio,
        map_event=_map_speech_event,
        stream_reset_timeout_secs=stream_reset_timeout_secs,
        max_resets=(60 * max_minutes) // stream_reset_timeout_secs,
    ).step(async_iter)
    return _google_speech_postprocessing(stream, include_events)


def google_speech_v1_step(
    async_iter: AsyncIterator[bytes],
    speech_async_client: SpeechAsyncClientV1,
    audio_format: AudioFormat,
    model: str = "latest_long",
    language_code: str = "en-US",
    include_events: bool = False,
    max_minutes: int = 9,
    stream_reset_timeout_secs=240,
) -> AsyncIterator[str]:
    """
    Data flow step for converting audio into text using Google Cloud Speech-to-Text V1 API.

    This function processes an asynchronous stream of audio bytes, using Google Cloud
    Speech-to-Text V1 service to convert the audio into text. It supports additional
    configuration such as specifying the model, language codes, and audio format.
    If 'include_events' is set to True, it also returns a stream of speech recognition
    events alongside the text (:class:`~voice_stream.events.SpeechStart` and :class:`~voice_stream.events.SpeechEnd`).

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator over audio data in bytes.
    speech_async_client : SpeechAsyncClient
        An instance of SpeechAsyncClientV1 for interacting with the Google Cloud Speech-to-Text API V1.
    audio_format : voice_stream.audio.AudioFormat
        The audio format of the input data.  This is required in V1.  Use the V2 API for auto-detection of formats.
    model : str, optional
        The model to be used by the recognizer. Default is "latest_long".
    language_code : str, optional
        The language code(s) for the recognizer. Default is "en-US".
    include_events : bool, optional
        If True, the function also returns a stream of speech recognition events. Default is False.
    max_minutes : int, optional
        The maximum number of minutes to process.  Default is 9.

    Returns
    -------
    Union[AsyncIterator[str], Tuple[AsyncIterator[str], AsyncIterator[voice_stream.events.BaseEvent]]]
        If `include_events` is False, returns an asynchronous iterator yielding recognized text from the audio stream.
        If `include-events` is True, returns a tuple with 2 iterators.  The first yields the recognized text, and the
        second contain speech events.

    Notes
    -----
    - The function breaks the audio stream into chunks, sends them to the Speech-to-Text
      API, and processes the responses to extract the transcript.
    - Speech recognition events include information like word timings and confidences.
    """
    initial_config = _initial_recognition_config_v1(
        include_events=include_events,
        model=model,
        language_code=language_code,
        audio_format=audio_format,
    )

    def map_audio(x):
        return StreamingRecognizeRequestV1(audio_content=x)

    stream = EndlessStream(
        speech_async_client,
        initial_config,
        audio_format=audio_format,
        map_audio=map_audio,
        map_event=_map_speech_event_v1,
        stream_reset_timeout_secs=stream_reset_timeout_secs,
        max_resets=(60 * max_minutes) // stream_reset_timeout_secs,
    ).step(async_iter)
    return _google_speech_postprocessing(
        stream,
        include_events,
    )


def _google_speech_postprocessing(
    async_iter: AsyncIterator[BaseEvent],
    include_events: bool,
):
    stream = async_iter
    # stream = log_step(stream, "SR EH out", lambda x: f"{format_current_task()}\n{x}")

    def extract_text(x):
        return x.text if isinstance(x, SpeechToTextResult) and len(x.text) > 0 else None

    if include_events:
        stream, events = partition_step(
            stream, lambda x: isinstance(x, SpeechToTextResult)
        )
        stream = map_step(stream, extract_text, ignore_none=True)
        return stream, events
    else:
        stream = map_step(stream, extract_text, ignore_none=True)
        return stream


def _get_transcript(result):
    if not result.results:
        logger.info("No results in google response")
        return ""
    if not result.results[0].alternatives:
        logger.info("No alternatives in google response")
        return ""
    return result.results[0].alternatives[0].transcript


def _resolve_google_audio_config(audio_format: GoogleAudioConfig, speaking_rate: float):
    if isinstance(audio_format, AudioConfig):
        return audio_format
    elif audio_format == AudioFormat.WAV_MULAW_8KHZ:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.MULAW,
            sample_rate_hertz=8000,
            speaking_rate=speaking_rate,
        )
    elif audio_format == AudioFormat.OGG_OPUS:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.OGG_OPUS, speaking_rate=speaking_rate
        )
    elif audio_format == AudioFormat.MP3:
        # noinspection PyTypeChecker
        return AudioConfig(
            audio_encoding=AudioEncoding.MP3, speaking_rate=speaking_rate
        )
    else:
        assert False, f"Unsupported audio format {audio_format}"


def _initial_recognition_config(
    include_events: bool,
    project: str,
    location: str,
    recognizer: str,
    model: str = "latest_long",
    language_codes: Union[str, list[str]] = ["en-US", "es-US"],
    audio_format: GoogleDecodingConfig = None,
):
    recognizer_path = (
        f"projects/{project}/locations/{location}/recognizers/{recognizer}"
    )
    auto_decoding, explicit_decoding = _resolve_audio_decoding(audio_format)

    # noinspection PyTypeChecker
    out = StreamingRecognizeRequest(
        recognizer=recognizer_path,
        streaming_config=StreamingRecognitionConfig(
            config=RecognitionConfig(
                auto_decoding_config=auto_decoding,
                explicit_decoding_config=explicit_decoding,
                features=RecognitionFeatures(enable_automatic_punctuation=True),
                model=model,
                language_codes=language_codes,
            ),
            streaming_features=StreamingRecognitionFeatures(
                interim_results=include_events,
                enable_voice_activity_events=True,
                # voice_activity_timeout=StreamingRecognitionFeatures.VoiceActivityTimeout(
                #     # speech_end_timeout
                # ),
            ),
        ),
    )
    # logger.debug(f"Initial recognition config: {out}")
    return out


def _initial_recognition_config_v1(
    include_events: bool,
    audio_format: AudioFormat,
    model: str = "latest_long",
    language_code: str = "en-US",
    use_enhanced: bool = True,
):
    if audio_format == AudioFormat.MP3:
        encoding = RecognitionConfigV1.AudioEncoding.MP3
        sample_rate_hertz = 44100
    elif audio_format == AudioFormat.WAV_MULAW_8KHZ:
        encoding = RecognitionConfigV1.AudioEncoding.MULAW
        sample_rate_hertz = 8000
    elif audio_format == AudioFormat.OGG_OPUS:
        encoding = RecognitionConfigV1.AudioEncoding.OGG_OPUS
        sample_rate_hertz = 48000
    elif audio_format == AudioFormat.WEBM_OPUS:
        encoding = RecognitionConfigV1.AudioEncoding.WEBM_OPUS
        sample_rate_hertz = 48000
    else:
        assert False, f"Unsupported audio format {audio_format}"

    # noinspection PyTypeChecker
    out = StreamingRecognizeRequestV1(
        streaming_config=StreamingRecognitionConfigV1(
            config=RecognitionConfigV1(
                encoding=encoding,
                sample_rate_hertz=sample_rate_hertz,
                model=model,
                language_code=language_code,
                enable_automatic_punctuation=True,
                use_enhanced=use_enhanced,
            ),
            interim_results=include_events,
            enable_voice_activity_events=True,
        ),
    )
    # logger.debug(f"Initial recognition config: {out}")
    return out


def _map_speech_event(input, time_offset: float = 0):
    if (
        input.speech_event_type
        == StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_BEGIN
    ):
        return SpeechStart(
            time_since_start=time_offset + input.speech_event_offset.total_seconds()
        )
    if (
        input.speech_event_type
        == StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_END
    ):
        return SpeechEnd(
            time_since_start=time_offset + input.speech_event_offset.total_seconds()
        )
    if input.results:
        transcript = _get_transcript(input)
        time_since = time_offset + input.results[-1].result_end_offset.total_seconds()
        if input.results[0].is_final:
            return SpeechToTextResult(text=transcript, time_since_start=time_since)
        else:
            return SpeechPartialResult(text=transcript, time_since_start=time_since)
    logger.warning(f"Unmapped speech event {input}")
    return None


def _map_speech_event_v1(input, time_offset: float = 0):
    if (
        input.speech_event_type
        == StreamingRecognizeResponseV1.SpeechEventType.SPEECH_ACTIVITY_BEGIN
    ):
        return SpeechStart(
            time_since_start=time_offset + input.speech_event_time.total_seconds()
        )
    if (
        input.speech_event_type
        == StreamingRecognizeResponseV1.SpeechEventType.SPEECH_ACTIVITY_END
    ):
        return SpeechEnd(
            time_since_start=time_offset + input.speech_event_time.total_seconds()
        )
    if input.results:
        transcript = _get_transcript(input)
        time_since = time_offset + input.results[-1].result_end_time.total_seconds()
        if input.results[0].is_final:
            return SpeechToTextResult(text=transcript, time_since_start=time_since)
        else:
            return SpeechPartialResult(text=transcript, time_since_start=time_since)
    if input.error:
        raise ValueError(
            f"Google Speech Recognition Error - Code: {input.error.code} Message: {input.error.message}"
        )
    logger.warning(f"Unmapped speech event {input}")
    return None



import asyncio
import logging
from typing import AsyncIterator, AsyncGenerator

import asyncstdlib

from voice_stream.types import (
    EndOfStreamMarker,
    QueueExceptionMarker,
    T,
    format_current_task,
)

logger = logging.getLogger(__name__)


class QueueWithException(asyncio.Queue):
    """
    A specialized asyncio.Queue that propagates exceptions to the consumers.

    This queue enhances the standard asyncio.Queue by handling exceptions. When an exception
    is set in the queue via `set_exception()`, it is propagated to the consumers either when they call `get()` or
    `get_nowait()`.

    Examples
    --------
    >>> queue = queue_source()
    >>> await queue.put(1)
    >>> await queue.put(2)
    >>> await queue.set_exception(RuntimeError("Test Error"))
    >>> await array_sink(queue)
    Caught exception: Test Error

    Notes
    -----
    - If an exception is set, it maintains its place in the queue.  If there are existing items in the queue, those will
      be returned in calls to `get()` until all items in the queue when `set_exception` was called have been exhausted.
    """

    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)
        self.exception = None

    async def enqueue_iterator(
        self,
        async_iter: AsyncIterator[T],
        include_end_of_stream: bool = True,
        propagate_exception: bool = True,
    ) -> AsyncGenerator[T, None]:
        """Enqueues items from an asynchronous iterator to the queue.

        If the asynchronous iterator is cancelled, no end of stream is appended.

        Parameters:
        -----------
            async_iter : AsyncIterator
                An asynchronous iterator whose items will be enqueued.
            include_end_of_stream : bool, optional
                Whether to include an end of stream marker after iteration is completed.
        """
        if not hasattr(async_iter, "__aiter__"):
            raise ValueError(f"Object {async_iter} is not an async iterator")
        try:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    await self.put(item)
            if include_end_of_stream:
                # logger.debug(f"End of stream in task {format_current_task()}")
                await self.put(EndOfStreamMarker)
        except Exception as e:
            if propagate_exception:
                # logger.debug(f"Queuing exception in task {format_current_task()} - {e}")
                await self.set_exception(e)
            else:
                raise e

    def _get(self):
        out = super()._get()
        if out == QueueExceptionMarker and self.exception:
            # logger.debug("Raising exception in queue")
            exception = self.exception
            self.exception = None
            raise exception
        return out

    async def set_exception(self, exception: Exception):
        """Sets an exception to be propagated to consumers.

        When an exception is set, the EndOfStreamMarker is enqueued and when that marker is removed from the queue,
        the exception will be raised.
        """
        if self.exception:
            raise ValueError(
                "set_exception called when the queue already had an exception."
            )
        self.exception = exception
        await self.put(QueueExceptionMarker)



"""
The `LangChain <https://www.langchain.com/>`_ is a framework for building LLM based applications.  VoiceStream allows
any LangChain runnable to be used to power a voice application.
"""
from __future__ import annotations

from typing import AsyncIterator, Callable, Any, Optional, Dict

import asyncstdlib
from langchain_core.memory import BaseMemory
from langchain_core.runnables import Runnable

from voice_stream.core import (
    fork_step,
    Output,
    map_step,
)
from voice_stream.speech_to_text import (
    filter_spurious_speech_start_events_step,
    SpeechStep,
)
from voice_stream.substreams import cancelable_substream_step
from voice_stream.text_to_speech import (
    TextToSpeechStep,
    tts_with_buffer_and_rate_limit_step,
)
from voice_stream.types import Input, SourceConvertable, to_source

GenericStepFunc = Callable[[AsyncIterator[Any]], AsyncIterator[Any]]


async def langchain_step(
    async_iter: AsyncIterator[Input],
    chain: Runnable[Input, Output],
    input_key: Optional[str] = None,
    config_key: Optional[str] = None,
    on_completion: SourceConvertable = None,
) -> AsyncIterator[Output]:
    """
    Data flow step that passes each input item to a LangChain runnable and streams the output.

    This step is used to call LLMs or run any other LangChain runnable.  It receives text
    items, processes them through a specified 'chain', and yields the resulting output
    tokens asynchronously.  If `input_key` is not specified, the input will be directly passed to the runnable.
    If `input_key` is specified, then the items coming from the source iterator must be dictionaries, and the `input_key`
    specifies which item should be passed as input to the runnable.  The `config_key` can also be used to pass configuration
    to the Runnable.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator that provides input text items.
    chain : Runnable[Input, Output]
        A Langchain Runnable that processes each input item.
    input_key : Optional[str], optional
        If specified, the item from the incoming dictionary that should be used as the input to the Runnable.
        If empty, then the incoming item will be passed directly.
    config_key : Optional[str], optional
        If specified, the item from the incoming dictionary that should be used as the config argument to the Runnable.
        Can only be specified if input_key is specified.
    on_completion : SourceConvertable, optional
        An optional source to be converted and iterated upon completion of processing each text item.  Can be used to
        signal the end of output if that isn't clear form the output of the chain itself.

    Returns
    ------=
    AsyncIterator[Output]
        An asynchronous iterator yielding the output from the LangChain Runnable.

    """
    # Note on cancelling - https://github.com/langchain-ai/langchain/issues/11959
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for text in owned_aiter:
            input = text[input_key] if input_key else text
            config = text[config_key] if config_key else None
            async for token in chain.astream(input, config=config):
                yield token
            source = to_source(on_completion)
            async with asyncstdlib.scoped_iter(source) as owned_source:
                async for item in owned_source:
                    yield item


def langchain_load_memory_step(
    async_iter: AsyncIterator[Dict],
    memory: BaseMemory,
) -> AsyncIterator[Dict]:
    """
    Data flow step that adds variables from LangChain memory into a dictionary.

    This step is used to insert variables from the memory into the current flow.  It expects a dictionary as input and
    returns a dictionary as output.  Used to prepare input for a langchain_step.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator that provides dictionaries.
    memory : BaseMemory
        A Langchain BaseMemory that contains the variables to insert

    Returns
    ------
    AsyncIterator[Dic]
        An asynchronous iterator yielding dictionaries with memory variables added.

    Notes
    -----
    - If the incoming dictionary contains any of the variables that are included in the history, they will override the
      values in the memory.
    """

    def load(x):
        m = memory.load_memory_variables(x)
        return {**m, **x}

    return map_step(async_iter, load)


def langchain_save_memory_step(
    async_iter: AsyncIterator[Dict],
    memory: BaseMemory,
    input_key="input",
    output_key="output",
) -> AsyncIterator[Dict]:
    """
    Data flow step that adds saves variables into LangChain memory.

    This step is used to save variables to memory into the current flow.  It expects a dictionary as input and
    returns the same dictionary as output.  It has a side effect of updating the memory.  Updating memory requires
    two dictionaries, one for input and one for output.  These are found by looking at the input_key and output_key
    values within the dictionary.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator that provides dictionaries.
    memory : BaseMemory
        A Langchain BaseMemory to update
    input_key : str
        Key within the input dictionary whose value will be used as the `input` parameter to save_context.
    output_key : str
        Key within the input dictionary whose value will be used as the `output` parameter to save_context.

    Returns
    ------
    AsyncIterator[Dic]
        An asynchronous iterator that is a copy of the input iterator.
    """

    def save(x):
        memory.save_context(inputs=x[input_key], outputs=x[output_key])
        return x

    return map_step(async_iter, lambda x: save(x))



import asyncio
import inspect
import json
import logging
import time
from asyncio import QueueEmpty, CancelledError
from typing import (
    AsyncIterator,
    Any,
    Tuple,
    Callable,
    Union,
    Optional,
    Type,
    Iterable,
    Coroutine,
    List,
)

import aiofiles
import asyncstdlib

from voice_stream.queue_with_exception import QueueWithException
from voice_stream.types import (
    to_tuple,
    resolve_awaitable_or_obj,
    T,
    Output,
    AwaitableOrObj,
    EndOfStreamMarker,
    is_async_iterator,
    to_source,
    format_current_task,
    format_stack_trace,
    background_task,
    cancel_with_confirmation,
)

logger = logging.getLogger(__name__)

# Streams are an abstraction on top of ASyncIterators.
# Sources are iterators.
# Sinks take an iterator.
# Steps take iterators and return iterators


def empty_source() -> AsyncIterator[T]:
    """
    Data flow source that returns an empty asynchronous iterator.

    This function provides a utility to create an empty asynchronous iterator. The returned iterator will immediately
    signal the end of iteration when iterated over, as it contains no elements.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that yields no items and immediately ends iteration.

    Examples
    --------
    >>> stream = empty_source()
    >>> done = await array_sink(stream)
    >>> assert done == []

    Notes
    -----
    - The generic type `T` in `AsyncIterator[T]` indicates that the function can theoretically return
      an iterator of any type, but since it's initialized with an empty list, no items of any type are yielded.
    """
    return array_source([])


def none_source() -> AsyncIterator[None]:
    """
    Data flow source that yields a single None value.

    This function is a simple utility for generating an asynchronous iterator that yields exactly one item,
    which is `None` before signaling the end of iteration.  This is useful when you want to send 'None' as a
    cancel signal to a substream function.

    Returns
    -------
    AsyncIterator[None]
        An asynchronous iterator that yields a single item, `None`.

    Examples
    --------
    >>> stream = empty_source()
    >>> done = await array_sink(stream)
    >>> assert done == [None]

    Notes
    -----
    - The type `T` in `AsyncIterator[T]` is a placeholder indicating the iterator can be of any type.
      In this specific case, the iterator yields items of type `NoneType`.

    See Also
    --------
    :func:`~voice_stream.interruptable_substream_step`
    :func:`~voice_stream.cancelable_substream_step`
    """
    return array_source([None])


async def empty_sink(async_iter: AsyncIterator[T]) -> None:
    """
    Data flow sink that performs no action on the received items.

    This function asynchronously iterates over all items from the provided iterator, effectively 'emptying' it.
    Each item is retrieved but not used. This is useful in cases where the useful work in a data flow is done as
    a side effect of a previous iterator.  Since data flows are "pull-based", each branch must end in some kind of sink.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator whose items will be consumed. The type `T` is generic and can represent
        any type of item that the iterator yields.

    Returns
    -------
    None
        This function does not return any value.

    Examples
    --------
    >>> # If your only goal is to log values and not do anything with them
    >>> # The data flow still needs to have a sink to run correctly.
    >>>
    >>> stream = array_source([1,2,3])
    >>> stream = log_step(stream, "Value")
    >>> await empty_sink(stream)
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for _ in owned_aiter:
            pass
    # logger.info("Empty sink finished")


async def single_source(item: T) -> AsyncIterator[T]:
    """
    Data flow source that yields a single specified item.

    This function generates an asynchronous iterator that yields exactly one item, the one passed as
    an argument.

    Parameters
    ----------
    item : T
        The item to be yielded by the asynchronous iterator. The generic type `T` allows for any type
        of item to be provided.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that yields the provided item once.

    Examples
    --------
    >>> stream = single_source("Hello, world")
    >>> out = await array_sink(stream)
    >>> assert out == ["Hello, world"]

    Notes
    -----
    -  Often used with a :func:`~voice_stream.concat_step` to add an object ot the beginning or end of
    a stream.
    """
    await asyncio.sleep(0)
    yield item


async def array_source(array: list[T]) -> AsyncIterator[T]:
    """
    Data flow source the yields items from a list.

    This function takes a list and converts it into an asynchronous iterator. Each element of the list
    is yielded one by one in an asynchronous fashion.

    Parameters
    ----------
    array : list[T]
        A list of items of type `T`. The elements of this list will be yielded by the asynchronous iterator.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that yields each element of the input list.

    Examples
    --------
    >>> stream = array_source([1,2,3])
    >>> stream = log_step(stream, "Value")
    >>> await empty_sink(stream)

    Notes
    -----
    - This source is useful when an existing list needs to be processed asynchronously.
    """
    for item in array:
        await asyncio.sleep(0)
        yield item


async def array_sink(async_iter: AsyncIterator[T]) -> list[T]:
    """
    Data flow sink that collects items into a list.

    This function asynchronously iterates over all items from the provided iterator and collects
    them into a list. This is useful for when you want to convert an asynchronous stream of data into a
    synchronous data structure like a list.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator from which items will be consumed. The generic type `T` can represent
        any type of item that the iterator yields.

    Returns
    -------
    list[T]
        A list containing all the items yielded by the asynchronous iterator.

    Examples
    --------
    >>> stream = array_source([1,2,3])
    >>> stream = log_step(stream, "Value")
    >>> out = await array_sink(stream)
    >>> assert out == [1,2,3]

    Notes
    -----
    - This sink is ideal for scenarios where the entirety of an asynchronous data stream needs to be
      collected and processed at once.
    - The sink does end up holding all the collected items in memory, which can cause memory issues with very large streams.
    """
    array = []
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            array.append(item)
        return array


class QueueAsyncIterator:
    """
    An asynchronous iterator that operates on its own queue.

    This class implements an async iterator which allows asynchronous iteration over
    queued items. It uses an asyncio.Queue for storing items and provides a `put` method
    to add items to this queue. The iterator retrieves items from the queue in the order
    they were added.

    Examples
    --------
    >>> queue = queue_source() # With no input arguments, this returns a QueueAsyncIterator.
    >>> stream = queue
    >>> done = array_sink()
    >>> await queue_iter.put(1)
    >>> await queue_iter.put(2)
    >>> await queue_iter.put(EndOfStreaMarker)
    >>> out = await done
    >>> assert out == [1, 2]
    """

    def __init__(self):
        self.queue = asyncio.Queue()
        self.iter = queue_source(self.queue)

    async def put(self, item):
        """Adds an item to the queue managed by this iterator.

        Parameters
        ----------
        item : any type
            The item to be added to the queue.
        """
        return await self.queue.put(item)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self.iter.__anext__()


def queue_source(
    queue: AwaitableOrObj[asyncio.Queue[T]] = None,
    cancel_event: asyncio.Event = None,
) -> Union[AsyncIterator[T], QueueAsyncIterator]:
    """
    Data flow source that yields items from an asyncio.Queue.

    This function returns an asynchronous iterator that consumes items from an asyncio.Queue.  The iteration continues
    until an `EndOfStreamMarker` is encountered in the queue, signaling the end of iteration.

    Parameters
    ----------
    queue : AwaitableOrObj[asyncio.Queue[T]], optional
        An instance of asyncio.Queue from which the items will be consumed. This can be an instance of the queue
        or an Awaitable that returns the queue, which can be useful if the queue isn't yet created.  This parameter is
        optional.  If not provided, the AsyncIterator will be a :func:`~voice_stream.QueueAsyncIterator` which allows
        items to be added to the queue via the `put` method.
    cancel_event : asyncio.Event
        An optional cancellation event that externally indicates that this sink should stop listening on the queue.
        This is useful if the queue will be reused for another operation.  Otherwise this event will sit around waiting
        for a new message.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator over the items in the queue.

    Examples
    --------
    >>> queue = asyncio.Queue()
    >>> queue.put(1)
    >>> queue.put(2)
    >>> queue.put(EndOfStreamMarker)
    >>> stream = queue_source()
    >>> done = await array_sink(stream)
    >>> assert done == [1,2]

    Notes
    -----
    - The function expects that the queue will be closed by putting `EndOfStreamMarker` into it.
    - If an Awaitable[Queue] is passed, this function will return immediately and the queue will be awaited when the iterator is started.
    - cancel_event is only allowed if a queue is passed in.  (It doesn't make sense otherwise.  If the queue is internally managed, it can't be reused).
    """
    if queue:
        if cancel_event:

            class WaitingIter:
                def __init__(self, queue, event):
                    self.queue = queue
                    self.event = event

                def __aiter__(self):
                    return self

                async def __anext__(self):
                    if self.event.is_set():
                        raise StopAsyncIteration
                    resolved_queue = await resolve_awaitable_or_obj(self.queue)
                    item_task = asyncio.create_task(resolved_queue.get())
                    stop_task = asyncio.create_task(self.event.wait())
                    done, pending = await asyncio.wait(
                        {item_task, stop_task}, return_when=asyncio.FIRST_COMPLETED
                    )
                    for task in pending:
                        await cancel_with_confirmation(task)
                    if item_task in done:
                        item = await item_task
                        if item == EndOfStreamMarker:
                            raise StopAsyncIteration
                        return item
                    else:
                        assert stop_task in done
                        raise StopAsyncIteration

            return WaitingIter(queue, cancel_event)

        else:

            async def gen():
                resolved_queue = await resolve_awaitable_or_obj(queue)
                while True:
                    try:
                        item = await resolved_queue.get()
                        if item == EndOfStreamMarker:
                            break
                    except asyncio.CancelledError:
                        # logger.debug("Queue iterator cancelled.")
                        raise
                    # logger.debug(f"Got {str(item)[:50]} from queue")
                    yield item
                # logger.debug("Queue exhausted.")

            return gen()
    else:
        return QueueAsyncIterator()


async def queue_sink(
    async_iter: AsyncIterator[T],
    queue: Optional[AwaitableOrObj[asyncio.Queue]] = None,
    end_of_stream: Any = EndOfStreamMarker,
    send_end_of_stream: bool = True,
    propagate_exceptions: bool = True,
) -> asyncio.Queue:
    """
    Data flow sink that writes items to a queue.

    This function takes an asynchronous iterator and writes each element it yields into a queue.
    If a queue is provided, it is used; otherwise, a new asyncio.Queue is created. Upon completion
    or exception it signals the end of the stream.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator from which elements are read.
    queue : Optional[AwaitableOrObj[asyncio.Queue]], optional
        An optional asyncio.Queue or an awaitable resulting in an asyncio.Queue.
        If not provided, a new :class:`~voice_stream:QueueWithException` is created.
    end_of_stream : The item to enqueue to indicate the stream has ended.  Defaults to EndOfStreamMarker
    send_end_of_stream : bool, optional
        If False, don't send any end of stream signal at all.
    propagate_exceptions : bool, optional
        If True, send exceptions to the queue.  Otherwise, raise them.  Defaults to True.

    Returns
    -------
    asyncio.Queue
        The queue to which the elements from async_iter are written.

    Examples
    --------
    >>> stream = array_source([1,2])
    >>> queue = await queue_sink()
    >>> assert len(queue) == 2
    >>> assert queue.get() == 1
    >>> assert queue.get() == 2
    >>> assert queue.empty()

    Notes
    --------
    - If the Queue has a `set_exception` method, that will be called when an exception is thrown by the iterator.  The
      :class:`~voice_stream:QueueWithException` has this method, and uses it to throw the exception on the next call to `queue.get`.
    """
    resolved_queue = None
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for message in owned_aiter:
                if resolved_queue is None:
                    if queue:
                        resolved_queue = await resolve_awaitable_or_obj(queue)
                    else:
                        resolved_queue = QueueWithException()
                # logger.debug(f"Put {str(message)[:50]} on queue")
                await resolved_queue.put(message)
    except Exception as e:
        if (
            propagate_exceptions
            and resolved_queue
            and hasattr(resolved_queue, "set_exception")
        ):
            await resolved_queue.set_exception(e)
        else:
            raise e
    if resolved_queue and send_end_of_stream:
        # logger.debug(f"Adding end of stream marker to queue {format_stack_trace(asyncio.current_task())}")
        await resolved_queue.put(end_of_stream)  # signal completion
    return resolved_queue


async def text_file_source(filename: str) -> AsyncIterator[str]:
    """
    Data flow source that yields the lines in a text file.

    This function asynchronously reads a text file line by line, yielding each line without the trailing newline
    character.

    Parameters
    ----------
    filename : str
        The path to the text file to be read.

    Returns
    -------
    AsyncIterator[str]
        An asynchronous iterator yielding lines from the file as strings.

    Raises
    ------
    FileNotFoundError
        If the file specified by filename does not exist.
    IOError
        If an I/O error occurs while opening or reading the file.

    Examples
    --------
    >>> stream = text_file_source("example.txt")
    >>> done = await text_file_sink(stream, "copy.txt")
    """
    async with aiofiles.open(filename, "rt") as f:
        async for line in f:
            line = line[:-1] if line[-1] == "\n" else line
            yield line


async def text_file_sink(
    async_iter: AsyncIterator[Any], filename: AwaitableOrObj[str]
) -> None:
    """
    Data flow sink that writes each element to a file, each as a new line.

    This function takes each element from the provided async iterator and writes it to the specified
    text file. Each element is converted to a string and written as a separate line. The filename
    can be either a string or an awaitable object that resolves to a string.

    Parameters
    ----------
    async_iter : AsyncIterator[Any]
        An asynchronous iterator whose elements are to be written to the file.
    filename : AwaitableOrObj[str]
        The path to the file where the data will be written. This can be a string or an
        awaitable object that yields a string.  The object will be awaited only after the first element is received
        from the iterator.

    Returns
    -------
    None
        This function does not return a value but completes when all elements have been
        written to the file.

    Examples
    --------
    >>> stream = text_file_source("example.txt")
    >>> done = await text_file_sink(stream, "copy.txt")
    """
    return await _file_sink(async_iter, filename, "wt", lambda x: f"{x}\n")


async def binary_file_source(
    filename: str, chunk_size: int = 4096
) -> AsyncIterator[bytes]:
    """
    Data flow source that yields chunks of bytes from a binary file.

    This function opens a binary file and asynchronously reads it in chunks of a specified size.
    It yields each chunk of bytes until the end of the file is reached. This is particularly useful
    for processing large binary files in an asynchronous manner.

    Parameters
    ----------
    filename : str
        The path to the binary file to be read.
    chunk_size : int, optional
        The number of bytes to read in each chunk. Defaults to 4096 bytes.  Pass 0 to read in the whole file at once.

    Returns
    -------
    AsyncIterator[bytes]
        An asynchronous iterator yielding chunks of bytes from the file.

    Raises
    ------
    FileNotFoundError
        If the file specified by filename does not exist.
    IOError
        If an I/O error occurs while opening or reading the file.

    Examples
    --------
    >>> stream = binary_file_source("example.bin")
    >>> stream = log_step(stream, "Bytes read", lambda x: len(x))
    >>> done = await binary_file_sink("copy.bin")
    Bytes read 4096
    Bytes read 4096
    Bytes read 1024  # Example output for a 9216-byte file
    """
    async with aiofiles.open(filename, "rb") as f:
        while True:
            chunk = await f.read(chunk_size)
            if not chunk:
                break
            yield chunk


async def binary_file_sink(
    async_iter: AsyncIterator[bytes], filename: AwaitableOrObj[str]
) -> None:
    """
    Data flow sink that writes chunks of bytes to a binary file.

    This function takes each block of bytes yielded by the provided async iterator and appends it to
    the specified binary file. The filename can be either a string or an awaitable object that resolves
    to a string.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator yielding blocks of bytes to be written to the file.
    filename : AwaitableOrObj[str]
        The path to the binary file where the data will be appended. This can be a string or an
        awaitable object that yields a string.

    Returns
    -------
    None
        This function does not return a value but completes when all blocks of bytes have been
        appended to the file.

    Raises
    ------
    IOError
        If an I/O error occurs while opening or writing to the file.

    Examples
    --------
    >>> stream = binary_file_source("example.bin")
    >>> stream = log_step(stream, "Bytes read", lambda x: len(x))
    >>> done = await binary_file_sink("copy.bin")
    Bytes read 4096
    Bytes read 4096
    Bytes read 1024  # Example output for a 9216-byte file

    Notes
    -----
    - If an awaitable is passed, it is not waited on until the first item has been retrieved from the iterator.
    """
    return await _file_sink(async_iter, filename, "wb", lambda x: x)


async def _file_sink(
    async_iter: AsyncIterator[T],
    filename: AwaitableOrObj[str],
    mode: str,
    prep_func: Callable[[T], Any],
) -> None:
    f = None
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for block in owned_aiter:
                # We wait to resolve the filename until we have a message to write
                if f is None:
                    filename = await resolve_awaitable_or_obj(filename)
                    f = await aiofiles.open(filename, mode)
                await f.write(prep_func(block))
    finally:
        if f is not None:
            await f.close()
            logger.debug(f"Closed {filename}")


def map_str_to_json_step(async_iter: AsyncIterator[str]) -> AsyncIterator[dict]:
    """
    Data flow step that parses JSON strings into dictionaries.

    Each string element from the async iterator is parsed as JSON and transformed into a dictionary.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator yielding strings, each expected to be a valid JSON text.

    Returns
    -------
    AsyncIterator[dict]
        An asynchronous iterator yielding dictionaries resulting from JSON parsing of each string element.

    Examples
    --------
    >>> stream = array_source([
    ...     '{"name": "Alice", "age": 30}',
    ...     '{"name": "Bob", "age": 25}',
    ... ])
    >>> stream = map_str_to_json_step(stream)
    >>> stream = map_step(stream, lambda x: x['age'])
    >>> await array_sink(stream)
    [30, 25]
    return map_step(async_iter, lambda x: json.loads(x))
    """
    return map_step(async_iter, lambda x: json.loads(x))


async def collect_dict_step(async_iter: AsyncIterator[dict]) -> AsyncIterator[dict]:
    """
    Data flow step that appends values to a dictionary until an empty value is passed.

    This step combines dictionary values coming out of an iterator.  It adds new keys to the output dictionary and
    appends values if possible (if not it replaces values).  It emits the current dictionary when an empty input
    is pulled form the incoming iterator.  Useful for combining incremental results coming from a streamed LangChain.

    Parameters
    ----------
    async_iter : AsyncIterator[Iterable[T]]
        An asynchronous iterator where each item is an iterable of elements.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator yielding combined dictionaries.

    Examples
    --------
    >>> stream = array_source([{'text':'Hello'}, {'text':' World!', 'confidence':0.75}, None])
    >>> stream = collect_dict_step(stream)
    >>> done = await array_sink(stream)
    >>> assert done == [{'text':"Hello World!", 'confidence': 0.75}]
    """

    def combine(x, y):
        try:
            return x + y
        except:
            return y

    buffer = {}
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if item:
                appended_vals = {
                    k: (v if k not in buffer else combine(buffer[k], v))
                    for k, v in item.items()
                }
                buffer = {**buffer, **appended_vals}
            else:
                yield buffer
                buffer = {}


async def delay_step(
    async_iter: AsyncIterator[T], delay_secs: float
) -> AsyncIterator[T]:
    """
    Data flow step that imposes a deplay on items coming in.

    This step simply adds a fixed delay to the items from the incoming iterator.  Mainly useful for testing.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator
    delay : float
        The number of seconds to sleep between each item

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator yielding the items coming from the input iterator with a delay.
    """

    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            try:
                await asyncio.sleep(delay_secs)
            except CancelledError as e:
                pass
            yield item


async def flatten_step(
    async_iter: AsyncIterator[Union[Iterable[T], AsyncIterator[T]]],
) -> AsyncIterator[T]:
    """
    Data flow step that flattens an iterator of iterators into a single iterator.

    This function takes an asynchronous iterator where each item is itself an async iterator or a regular iterable
    (like a list or tuple) and flattens it. This means it iterates over each element of these iterables in order, yielding them
    individually. It's useful for converting a stream of iterables into a flat stream of elements.

    Parameters
    ----------
    async_iter : AsyncIterator[Iterable[T]]
        An asynchronous iterator where each item is an iterable of elements.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator yielding individual elements from each iterable in the input async iterator.

    Examples
    --------
    >>> stream = array_source([[1,2], [3,4]])
    >>> stream = flatten_step(stream)
    >>> done = await array_sink(stream)
    >>> assert done == [1, 2, 3, 4]
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if is_async_iterator(item):
                async with asyncstdlib.scoped_iter(item) as owned_subiter:
                    async for subitem in owned_subiter:
                        yield subitem
            else:
                for subitem in item:
                    yield subitem


def async_init_step(
    async_iter: AsyncIterator[T],
    f: Callable[[AsyncIterator[T]], Coroutine],
    num_outputs: int = 1,
) -> AsyncIterator[Output]:
    """
    Data flow step that asyncronously initializes a step.

    This step takes an AsyncIterator and a lambda that takes the iterator and returns an Awaitable.  It eagerly pulls
    the first item from the iterator and separately calls the lambda and awaits the result in a separate task.  This
    is useful when you ahve a step that requires an async call to initialize (which we generally avoid).  It's also
    useful when the initialization depends on a value produced from earlier in the data flow.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An input asynchronous iterator.
    f : Callable[[AsyncIterator[T]], Coroutine]
        A coroutine function that takes the input async iterator and returns another async iterator
    num_outputs : int, optional
        The number of output async iterators to expect from `f`. Defaults to 1.  Use this if the step you are
        asynchronously initializing returns more than one iterator, such as a :class:`~voice_stream:fork_step`.
        num_outputs can be 0 in the case where the asynchronous initialization is for a sink.

    Returns
    -------
    AsyncIterator[Output] or Tuple[AsyncIterator[Output], ...]
        if `num_outputs` is 1, then returns an AsyncIterator
        otherwise, returns a tuple of AsyncIterators the same length as `num_outputs`

    Examples
    --------
    # Copy a file, generating a name based on the first line of the file.
    >>> stream = text_file_source("example.txt")
    >>> stream, name = extract_value_step(stream, lambda x: x)
    >>> done = await async_init_step(stream, lambda x: text_file_sink(x, resolve_awaitable_or_obj(name)), num_outputs=0)
    """

    class AsyncInitInputIterator:
        """Wraps the input iterator and eagerly gets the first item.  This pulls data through the stream in case our
        async_init function is waiting on it."""

        def __init__(self):
            self.iter = async_iter.__aiter__()
            # Using create_task here because any exception in the result will get through in __anext__
            self.first_item_task = asyncio.create_task(self.iter.__anext__())

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.first_item_task:
                out = self.first_item_task
                self.first_item_task = None
                # logger.debug(f"Waiting for first item to async init step {format_current_task()}")
                return await out
            else:
                return await self.iter.__anext__()

    class AsyncInitOutputIterator:
        """Wraps the output and doesn't create the real iterator until the first item is requested."""

        def __init__(self, init_event):
            self.iter = None
            self.init_event = init_event
            self.init_exception: Optional[Exception] = None

        def __aiter__(self):
            return self

        async def __anext__(self):
            if not self.iter:
                await self.init_event.wait()
                if self.init_exception:
                    raise self.init_exception
                # logger.debug(f"Async init completed.  Waiting for next step in {format_current_task()}.")
            return await self.iter.__anext__()

    init_event = asyncio.Event()
    outputs = [AsyncInitOutputIterator(init_event) for _ in range(num_outputs)]

    async def do_init():
        try:
            input = AsyncInitInputIterator()
            # logger.debug(f"Beginning async init in {format_current_task()}.")
            step = await f(input)
            # logger.debug(f"Completed async init in {format_current_task()}.")
            for output, iterable in zip(outputs, to_tuple(step)):
                output.iter = iterable.__aiter__()
        except Exception as e:
            # logger.exception("Exception in async init.")
            for output in outputs:
                output.init_exception = e
        finally:
            init_event.set()

    background_task(do_init())

    if len(outputs) == 1:
        return outputs[0]
    else:
        return outputs


async def abort_step(
    async_iter: AsyncIterator[T], event: asyncio.Event
) -> AsyncIterator[T]:
    """
    Data flow step that ends when an event is set.

    This step passes along items from the input iterator until an event is set, and then it ends.  Useful for cleanup tasks.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An input asynchronous iterator.
    event : asyncio.Event
        The event that signals that iteration should end.

    Returns
    -------
    AsyncIterator[T]
        An iterator returning the input values.

    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if event.is_set():
                return
            yield item


def extract_value_step(
    async_iter: AsyncIterator[T],
    value: Callable[[T], Any],
    condition: Optional[Callable[[T], bool]] = None,
) -> Tuple[AsyncIterator[T], asyncio.Future[Any]]:
    """
    Extracts a value from an async iterator based on a condition, returning the iterator and a future.

    This function processes an asynchronous iterator, applying a given condition to each item. The first time
    an item meets the condition, a specified value extraction function is applied to it, and the result
    is set in an asyncio.Future. The function returns an async iterator and the future containing the extracted value.
    The output iterator produces all the same elements that go into it.  If the condition is not provided, the value
    function is applied to the first item.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An input asynchronous iterator.
    value : Callable[[T], Any]
        A function that extracts a value from an item in the iterator.
    condition : Optional[Callable[[T], bool]], optional
        A function that returns True if the value should be extracted from the item.
        If None, the value is extracted from the first item.

    Returns
    -------
    Tuple[AsyncIterator[T], asyncio.Future[Any]]
        The modified async iterator and a future containing the extracted value.

    Examples
    --------
    # Copy a file, generating a name based on the first line of the file.
    >>> stream = text_file_source("example.txt")
    >>> stream, name = extract_value_step(stream, lambda x: x)
    >>> done = await async_init_step(stream, lambda x: text_file_sink(x, resolve_awaitable_or_obj(name)), num_outputs=0)
    """
    loop = asyncio.get_running_loop()
    fut = loop.create_future()

    async def chain():
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for item in owned_aiter:
                if not fut.done() and (condition is None or condition(item)):
                    fut.set_result(value(item))
                yield item

    return chain(), fut


async def recover_exception_step(
    async_iter: AsyncIterator[T],
    exception_type: Type[BaseException],
    exception_handler: Callable[[BaseException], Any],
) -> AsyncIterator[T]:
    """
    Data flow step that recovers from an exception in a previous step.

    This function takes an async iterator and a specified exception type. If an exception with the
    specified type occurs during iteration, it is caught and passed to the provided exception
    handler function. The result depends on what is returned by the exception handler:
    * If it returns an AsyncIterator, that will be yielded and this step will end when the iterator ends.
    * If it returns None, this step will immediately end.
    * If it returns anything else, that object will be yielded and this step will end.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator to be wrapped.
    exception_type : Type[BaseException]
        The type of exception to be caught and handled.
    exception_handler : Callable[[BaseException], Any]
        A function to handle the caught exception.  Returns an optional value that will be converted to an AsyncIterator.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that handles exceptions with the specified type.

    Examples
    --------
    >>> def gen():
    ...    yield 1
    ...    yield 2
    ...    raise ValueError("No more")
    >>> stream = gen()
    >>> stream = recover_exeption_step(stream, ValueError, lambda x: x.args[0])
    >>> done = await array_sink(stream)
    >>> assert done == [1,2,"No more"]

    Notes
    -----
    - Once the source iterator throws the exception, it will not be accessed again.
    """
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for item in owned_aiter:
                yield item
    except exception_type as e:
        eh = await resolve_awaitable_or_obj(exception_handler(e))
        handler_iter = to_source(eh)
        async with asyncstdlib.scoped_iter(handler_iter) as owned_aiter:
            async for item in owned_aiter:
                yield item


async def log_step(
    async_iter: AsyncIterator[T],
    name: str,
    formatter: Callable[[T], Any] = lambda x: x,
    every_nth_message: Optional[int] = None,
) -> AsyncIterator[T]:
    """
    Data flow step that prints using the Python logger.

    This function takes an async iterator and logs each item that passes through it. The logging
    includes a specified name and uses an optional formatter function to convert items to a log-friendly format.
    This is useful for debugging or monitoring the contents of an data flow.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator whose items are to be logged.
    name : str
        A name to prepend to each logged message for identification.
    formatter : Callable[[T], Any], optional
        A function to format each item before logging. Defaults to a function that returns the item unchanged.
    every_nth_message : Optional[int]
        Indicates that the log should only happen every nth message.  The first message will always cause a log message,
        and then every nth message after the first.  For example, if the value is 10, then the 1st, 11th, 22nd, etc.
        messages will be logged.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that logs each item and then yields it.

    Examples
    --------
    >>> stream = array_source([1,2,3])
    >>> stream = log_step(stream, "Value squared", lambda x: x*x)
    >>> await empty_sink()
    Value squared: 1
    Value squared: 4
    Value squared: 9
    """
    count = 0
    log = True
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if every_nth_message and every_nth_message > 1:
                count += 1
                log = count % every_nth_message == 1
            if log:
                formatted = formatter(item)
                logger.info(f"{name}: {formatted}")
            yield item


async def count_step(async_iter: AsyncIterator[T], name: str) -> AsyncIterator[T]:
    """
    Data flow step that counts the number of items.

    This function takes an async iterator and counts the number of items that pass through it.
    After the iteration completes, it logs the total count of items. This is useful for monitoring or debugging the
    flow of data through async streamlines.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator whose items are to be counted.
    name : str
        A name to use in the log message for identifying the count source.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that counts each item and then yields it.

    Examples
    --------
    >>> stream = array_source([1,2,3])
    >>> stream = count_step(stream, "Value")
    >>> await empty_sink()
    Value count: 3
    """
    counter = 0
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            counter += 1
            yield item
    logger.info(f"{name} count: {counter}")


async def filter_step(
    async_iter: AsyncIterator[T], condition: Callable[[T], bool]
) -> AsyncIterator[T]:
    """
    Data flow step that filters items based on a specified condition.

    This function wraps an async iterator and yields only those items that satisfy a given condition.
    It is analogous to the built-in `filter` function but for asynchronous iterators. Each item from
    the input async iterator is passed to the `condition` function, and only items for which this
    function returns `True` are yielded.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator whose items are to be filtered.
    condition : Callable[[T], bool]
        A function that evaluates each item in the iterator. If this function returns `True`,
        the item is yielded; otherwise, it is skipped.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator yielding only the items that satisfy the condition.

    Examples
    --------
    >>> stream = array_source(range(4))
    >>> stream = filter_step(stream, lambda x: x % 2 == 0)
    >>> done = await array_sink(stream)
    >>>
    # Output: 0, 2
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if condition(item):
                yield item


async def map_step(
    async_iter: AsyncIterator[T],
    func: Callable[[T], Output],
    ignore_none: Optional[bool] = False,
) -> AsyncIterator[Output]:
    """
    Data flow step that transforms items using a mapping function.

    This function applies either a synchronous or asynchronous mapping function to each item in the
    input async iterator. If `ignore_none` is set to True, any items that are transformed to `None`
    are not yielded. This feature allows the function to perform both transformation and filtering in a single step.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator whose items are to be transformed.
    func : Callable[[T], Output]
        A mapping function to apply to each item. This can be either a synchronous or asynchronous function.
    ignore_none : Optional[bool], default False
        If True, items that are transformed to None by `func` are not yielded.

    Returns
    -------
    AsyncIterator[Output]
        An asynchronous iterator yielding transformed items, optionally skipping None values.

    Examples
    --------
    >>> stream = array_source(range(4))
    >>> stream = map_step(stream, lambda x: x + 2)
    >>> done = await array_sink(stream)
    >>> assert done == [2,3,4,5]
    """
    is_async = inspect.iscoroutinefunction(func)
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            v = func(item)
            if is_async:
                v = await v
            if (not ignore_none) or (v is not None):
                yield v


async def timeout_step(
    async_iter: AsyncIterator[bytes], timeout: float
) -> AsyncIterator[bytes]:
    start = time.perf_counter()
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            now = time.perf_counter()
            if (now - start) > timeout:
                logger.debug("Raising TimeoutError from timeout_step")
                raise TimeoutError()
            yield item


async def chunk_bytes_step(
    async_iter: AsyncIterator[bytes], chunk_size: AwaitableOrObj[int]
) -> AsyncIterator[bytes]:
    """
    Data flow step that breaks long byte sequences from an asynchronous iterator into fixed-size chunks.

    This function takes an asynchronous iterator yielding byte sequences and a chunk size,
    then yields these byte sequences in chunks of the specified size. The chunk size can
    be provided either as a direct integer value or as an awaitable object that resolves
    to an integer. This is useful for processing large byte sequences in manageable sizes.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator that yields byte sequences.
    chunk_size : AwaitableOrObj[int]
        The size of each chunk as an integer, or an awaitable object that yields an integer.

    Returns
    -------
    AsyncIterator[bytes]
        An asynchronous iterator yielding byte chunks of the specified size.

    Examples
    --------
    >>> stream = array_source([b'HelloWorld!', b'PythonAsyncIO']):
    >>> stream = chunk_bytes_step(stream, chunk_size=5)
    >>> done = await array_sink(stream)
    >>> assert done == [b'Hello', b'World', b'!', b'Pytho', b'nAsyn', b'cIO']
    """
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            resolved_chunk_size = await resolve_awaitable_or_obj(chunk_size)
            for c in _chunk_bytes(item, resolved_chunk_size):
                yield c


def _chunk_bytes(item, chunk_size):
    for i in range(0, len(item), chunk_size):
        data = item[i : i + chunk_size]
        # logger.debug(f"Chunked {len(data)} bytes")
        yield data


async def min_size_bytes_step(
    async_iter: AsyncIterator[bytes], min_size: AwaitableOrObj[int]
) -> AsyncIterator[bytes]:
    """
    Data flow steps that aggregates byte sequences to ensure a minimum size for each emitted chunk.

    This function takes an asynchronous iterator yielding byte sequences and a minimum size. It aggregates
    the byte sequences and yields them only when they reach or exceed the specified minimum size. The minimum
    size can be provided either as a direct integer value or as an awaitable object that resolves to an
    integer. This is useful for ensuring that chunks of data meet a size requirement before processing.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator that yields byte sequences.
    min_size : AwaitableOrObj[int]
        The minimum size of each chunk as an integer, or an awaitable object that yields an integer.

    Returns
    -------
    AsyncIterator[bytes]
        An asynchronous iterator yielding byte chunks that are at least the specified minimum size.

    Examples
    --------
    >>> stream = array_source([b'Hello', b'World', b'PythonAsyncIO', b'End')
    >>> stream = min_size_bytes_step(stream, min_size=10)
    >>> done = await array_sink(stream)
    >>> assert done == [
    ...     b'HelloWorld',
    ...     b'PythonAsyncIO',
    ...     b'End',
    ... ]
    """
    buffer = b""
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            combined = buffer + item
            if len(combined) >= await resolve_awaitable_or_obj(min_size):
                yield combined
                buffer = b""
            else:
                buffer = combined
        if len(buffer) > 0:
            yield buffer


async def concat_step(*async_iters: List[AsyncIterator[T]]) -> AsyncIterator[T]:
    """
    Data flow step that combines multiple streams by concatenating them.

    This function takes a list of asynchronous iterators and concatenates their elements
    into a single async iterator. It iterates over each input iterator in the order they
    are provided, yielding all items from each before moving on to the next.

    Parameters
    ----------
    async_iters : List[AsyncIterator[T]]
        A list of asynchronous iterators to be concatenated.

    Returns
    -------
    AsyncIterator[T]
        A single asynchronous iterator yielding items from all provided iterators in sequence.

    Examples
    --------
    >>> stream1 = array_source([1,2])
    >>> stream2 = array_source([3,4])
    >>> stream = concat_step(stream1, stream2)
    >>> done = await array_sink(stream)
    >>> assert done == [1,2,3,4]
    """
    for async_iter in async_iters:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for item in owned_aiter:
                yield item


async def merge_step(*async_iters: list[AsyncIterator[T]]) -> AsyncIterator[T]:
    """
    Data flow step that merges multiple streams into one, consuming from all iterators concurrently.

    Unlike concatenation which consumes one iterator at a time, `merge_step` interleaves items from all
    provided iterators as they become available. The merged iterator continues until all input iterators
    are exhausted. If an exception occurs in any of the iterators, all others are canceled.

    Parameters
    ----------
    async_iters : list[AsyncIterator[T]]
        A list of asynchronous iterators to be merged.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator yielding items from all input iterators in the order they become available.

    Examples
    --------
    >>> async def async_iter1():
    ...     for i in [1,2,3]:
    ...         await asyncio.sleep(0.3)
    ...         yield i
    >>> async def async_iter2():
    ...     for i in [4,5,6]:
    ...         await asyncio.sleep(0.5)
    ...         yield i
    >>> stream1 = async_iter1()
    >>> stream2 = async_iter2()
    >>> stream = merge_step(stream1, stream2)
    >>> done = await array_sink(stream)
    >>> assert done == [1,4,2,3,5,6]
    # Output order is not guaranteed.
    """
    # TODO Implement with backpressure.
    queue = QueueWithException()
    tasks = [background_task(queue.enqueue_iterator(it)) for it in async_iters]

    try:
        completed = 0
        while completed < len(async_iters):
            item = await queue.get()
            if item == EndOfStreamMarker:
                completed += 1
            else:
                yield item
    finally:
        for task in tasks:
            task.cancel()  # Cancel all tasks


async def merge_as_dict_step(
    dict_iters: dict[str, AsyncIterator[Any]]
) -> AsyncIterator[dict]:
    """
    Data flow step that merges multiple streams into a single one that yields dictionaries.

    Each iterator in the input dictionary is associated with a key. The function concurrently consumes
    items from all iterators and yields dictionaries. Each dictionary contains the most recent item
    from each iterator, keyed by the respective iterator's key. The output frequency is driven by the
    first iterator specified in the input dictionary.

    Parameters
    ----------
    dict_iters : dict[str, AsyncIterator[Any]]
        A dictionary mapping keys to asynchronous iterators. The first iterator in the dictionary
        drives the output frequency.

    Returns
    -------
    AsyncIterator[dict]
        An asynchronous iterator yielding dictionaries. Each dictionary combines the most recent items
        from each input iterator under their corresponding keys.

    Raises
    ------
    ValueError
        If `dict_iters` is empty.

    Examples
    --------
    >>> async def async_iter1():
    ...     for i in [1,2,3]:
    ...         await asyncio.sleep(0.3)
    ...         yield i
    >>> async def async_iter2():
    ...     for i in [4,5,6]:
    ...         await asyncio.sleep(0.5)
    ...         yield i
    >>> stream1 = async_iter1()
    >>> stream2 = async_iter2()
    >>> stream = merge_as_dict_step({'a':stream1, 'b':stream2})
    >>> done = await array_sink(stream)
    >>> assert done == [
    ...     {'a':1},
    ...     {'a':2, 'b':4},
    ...     {'a':3, 'b':4},
    ... ]
    """
    # logger.debug(f"merge_as_dict_step")
    if not dict_iters:
        raise ValueError("merge_as_dict_step requires at least on iterator")
    key_iter = iter(dict_iters)
    first_key = next(key_iter)
    first_iter = dict_iters[first_key]
    secondary_iters = {k: dict_iters[k] for k in key_iter}
    values = {}

    async def consume_iter(key, it):
        # logger.debug(f"Starting consumer for {key}")
        async with asyncstdlib.scoped_iter(it) as owned_it:
            async for item in owned_it:
                # logger.debug(f"Consumed {item} for {key}")
                values[key] = item

    tasks = [
        background_task(consume_iter(k, iter)) for k, iter in secondary_iters.items()
    ]
    async for item in first_iter:
        # logger.debug(f"Merging dict for {item}")
        yield {first_key: item, **values}
    for task in tasks:
        task.cancel()


def partition_step(
    async_iter: AsyncIterator[T], condition: Callable[[T], bool]
) -> Tuple[AsyncIterator[T], AsyncIterator[T]]:
    """
    Data flow step that partitions items into two output streams based on a specified condition.

    This function divides the items from the input async iterator into two separate iterators: one
    for items where the condition evaluates to True, and another for items where the condition is False.
    If an exception occurs during iteration, it is propagated to both output iterators.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator whose items are to be partitioned.
    condition : Callable[[T], bool]
        A function that evaluates each item in the iterator. Returns True if the item should go to the
        first iterator, and False for the second.

    Returns
    -------
    Tuple[AsyncIterator[T], AsyncIterator[T]]
        A tuple of two asynchronous iterators: the first yields items where the condition is True,
        and the second yields items where the condition is False.

    Examples
    --------
    >>> stream = array_source(range(4))
    >>> true_stream, false_stream = partition_step(stream, lambda x: x % 2 == 0)
    >>> true_done = array_sink(true_stream)
    >>> false_done = array_sink(false_stream)
    >>> true_ret, false_ret = await asyncio.gather(true_done, false_done)
    >>> assert true_ret == [0,2]
    >>> assert false_ret == [1,3]
    """
    true_queue = QueueWithException()
    false_queue = QueueWithException()

    async def distribute():
        try:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    # logger.debug(f"Distributing {item}")
                    if condition(item):
                        await true_queue.put(item)
                    else:
                        await false_queue.put(item)
                    # logger.debug(f"Queue sizes: {true_queue.qsize()} {false_queue.qsize()}")
        except Exception as e:
            # logger.debug(f"Exception while queueing")
            await true_queue.set_exception(e)
            await false_queue.set_exception(e)
        await true_queue.put(EndOfStreamMarker)  # Signal end of iteration
        await false_queue.put(EndOfStreamMarker)

    distribution_task = background_task(distribute())
    true_iterator = queue_source(true_queue)
    false_iterator = queue_source(false_queue)

    return true_iterator, false_iterator


def fork_step(
    async_iter: AsyncIterator[T],
    pull_from_all: bool = False,
) -> Tuple[AsyncIterator[T], AsyncIterator[T]]:
    """
    Data flow step that splips a stream into two, creating a 'fork' in the stream.

    This function takes an async iterator and divides its output into two separate async iterators.
    If `pull_from_all` is set to False (default), the consumption rate of items is determined by the
    rate of consumption of the first returned iterator; the second iterator's output rate will match
    the first. If `pull_from_all` is True, both iterators will consume items independently as quickly
    as possible.

    If pull_from_all is False, exceptions will only propagate to the first iterator. If True, exceptions
    will propagate to both iterators.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator to be forked.
    pull_from_all : bool, default False
        If true, both forks will pull form the incoming iterator.  If False, the second iterator will only
        receive data as fast as the first pulls.

    Returns
    -------
    Tuple[AsyncIterator[T], AsyncIterator[T]]
        A tuple of two asynchronous iterators representing the forked output.

    Examples
    --------
    >>> source = array_source(range(2))
    >>> a, b = fork_step(source)
    >>> a = await array_sink(a)
    >>> b = await array_sink(b)
    >>> assert a == [0, 1]
    >>> assert b == [0, 1]

    Notes
    -----
    - With pull_from_all set to False, the rate of iteration in the first iterator controls the overall
      pace of item consumption from the original iterator.
    - With pull_from_all set to True, items are consumed from the original iterator as quickly as possible,
      and both iterators receive items independently.
    """

    def no_pull_from_all():
        right_queue = asyncio.Queue()
        right_iterator = queue_source(right_queue)

        async def consume_and_queue() -> AsyncIterator[T]:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    await right_queue.put(item)
                    yield item
            await right_queue.put(EndOfStreamMarker)

        left_iterator = consume_and_queue()
        return left_iterator, right_iterator

    def with_pull_from_all():
        left_queue = QueueWithException()
        right_queue = QueueWithException()

        async def distribute():
            try:
                async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                    async for item in owned_aiter:
                        await left_queue.put(item)
                        await right_queue.put(item)
            except Exception as e:
                await left_queue.set_exception(e)
                await right_queue.set_exception(e)
            await left_queue.put(EndOfStreamMarker)  # Signal end of iteration
            await right_queue.put(EndOfStreamMarker)

        background_task(distribute())
        left_iterator = queue_source(left_queue)
        right_iterator = queue_source(right_queue)

        return left_iterator, right_iterator

    iterators = with_pull_from_all() if pull_from_all else no_pull_from_all()
    return iterators


def buffer_step(
    async_iter: AsyncIterator[T], join_func: Optional[Callable[[List[T]], T]] = None
) -> AsyncIterator[T]:
    """
    Data flow step that buffers and aggregates items them before yielding.

    This function acts as a buffer.  It pulls items from the incoming iterator as fast as they can come in, and
    then outputs aggreged items only as fast as the output iterator can pull them.  The aggregation is performed using a
    specified join function.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        The asynchronous iterator to buffer from.
    join_func : Callable[[List[T]], T], optional
        A function that takes a list of items and returns an aggregated result.  This is useful for buffering strings or
        byte arrays.  If not specified, then the items are returned exactly as they are put on the buffer.


    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator that yields aggregated items.

    Examples
    --------
    >>> async def rate_limit_sink(ai):
    ...     out = []
    ...     async for item in ai:
    ...         await asyncio.sleep(0.1)
    ...         out.append(item)
    >>> stream = array_source(['a','b','c','d'])
    >>> stream = buffer_step(stream, lambda x:  ''.join(x))
    >>> done = await rate_limit_sink(stream)
    >>> assert done == ['abcd']
    """
    queue = QueueWithException()
    task = background_task(queue.enqueue_iterator(async_iter))

    async def make_iterator() -> AsyncIterator[T]:
        try:
            ongoing = True
            while ongoing:
                # Use a blocking get() to ensure there's at least one chunk of
                # data, and stop iteration if the chunk is None, indicating the
                # end of the audio stream.
                item = await queue.get()
                if item == EndOfStreamMarker:
                    # logger.debug(f"End of queue")
                    return
                if join_func:
                    data = [item]
                    # Now consume whatever other data's still buffered.
                    while True:
                        try:
                            chunk = queue.get_nowait()
                            if chunk == EndOfStreamMarker:
                                # logger.debug(f"Got None in nowait")
                                ongoing = False
                                break
                            data.append(chunk)
                        except QueueEmpty:
                            # logger.debug(f"Got QueueEmpty in nowait")
                            break
                    out = join_func(data)
                    # logger.debug(f"Joining {data} {out}")
                    yield out
                else:
                    yield item
        finally:
            task.cancel()

    return make_iterator()


def byte_buffer_step(async_iter: AsyncIterator[bytes]) -> AsyncIterator[bytes]:
    """
    Data flow step that buffers and aggregates byte sequences to match producer and consumer rates.

    This function acts as a buffer specifically for byte streams.  It pulls bytes from the incoming iterator as fast
     as they can come in, and then outputs aggreged byte buffers only as fast as the output iterator can pull them.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator yielding byte sequences.

    Returns
    -------
    AsyncIterator[bytes]
        An asynchronous iterator yielding concatenated byte sequences.

    Examples
    --------
    >>> async def rate_limit_sink(ai):
    ...     out = []
    ...     async for item in ai:
    ...         await asyncio.sleep(0.5)
    ...         out.append(item)
    >>> stream = array_source([b'a',b'b',b'c',b'd'])
    >>> stream = byte_buffer_step(stream)
    >>> done = await rate_limit_sink(stream)
    >>> assert done == [b'abcd']
    """
    return buffer_step(async_iter, lambda x: b"".join(x))


def str_buffer_step(async_iter: AsyncIterator[str]) -> AsyncIterator[str]:
    """
    Data flow step that buffers and aggregates text strings to match producer and consumer rates.

    This function acts as a buffer specifically for text string.  It pulls text from the incoming iterator as fast
     as it can come in, and then outputs concatenated text only as fast as the output iterator can pull.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator yielding text.

    Returns
    -------
    AsyncIterator[str]
        An asynchronous iterator yielding concatenated text.

    Examples
    --------
    >>> async def rate_limit_sink(ai):
    ...     out = []
    ...     async for item in ai:
    ...         await asyncio.sleep(0.5)
    ...         out.append(item)
    >>> stream = array_source(['a','b','c','d'])
    >>> stream = str_buffer_step(stream)
    >>> done = await rate_limit_sink(stream)
    >>> assert done == ['abcd']
    """
    return buffer_step(async_iter, lambda x: "".join(x))



import asyncio
import logging
import string
import time
from typing import AsyncIterator, Tuple, Callable, Union

import asyncstdlib
from pydantic import BaseModel

from voice_stream.audio import AudioFormat, get_audio_length, audio_rate_limit_step
from voice_stream.core import (
    fork_step,
    map_step,
    extract_value_step,
    AwaitableOrObj,
    str_buffer_step,
    filter_step,
)
from voice_stream.events import TimedText
from voice_stream.types import resolve_awaitable_or_obj

logger = logging.getLogger(__name__)


class AudioWithText(BaseModel):
    """
    Representing audio data along with its corresponding text.  Used as the output of a text-to-speech step.
    """

    audio: bytes
    """The audio data."""

    text: str
    """The corresponding text for the audio."""

    audio_format: AudioFormat
    """The format of the audio."""


TextToSpeechStep = Callable[[AsyncIterator[str]], AsyncIterator[AudioWithText]]


def buffer_tts_text_step(async_iter: AsyncIterator[str]) -> AsyncIterator[str]:
    """
    Data flow step that buffers text for input to Text-to-Speech (TTS).

    When performing realtime TTS off of a token stream, there is a tradeoff.  You want to produce audio as soon as
    possible after receiving the first token, but passing longer utterances to a TTS engine produces more natural sounding
    speech.  This step buffers tokens to achieve a good balance by:

    1. Waiting for punctuation: Waits for some form of punctuation in the token stream to indicate the end of a phrase
       and then passes the full phrase.
    2. Buffers tokens received between pulls.  This step eagerly pulls tokens from the incoming iterator and buffers them.
       This way, any new tokens that come in while the TTS system is generating the current utterance are grouped together.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator that provides a stream of text.

    Returns
    -------
    AsyncIterator[str]
        The buffered TTS text.

    Example
    -------
    >>> stream = array_source(["Hello", " world!", " How", " are", " you", " today", ""])
    >>> stream = buffer_tts_text_step(stream)
    >>> out = await array_sink(stream)
    >>> assert out == ["Hello world!", " How are you today"]

    Note
    -------
    - An incoming empty string indicates an end of response and will cause the buffer to be flushed.

    """
    stream = wait_for_punctuation_step(async_iter)
    stream = str_buffer_step(stream)
    stream = filter_step(stream, lambda x: len(x) > 0)
    return stream


def tts_with_buffer_and_rate_limit_step(
    async_iter: AsyncIterator[str],
    tts_step: TextToSpeechStep,
    buffer_seconds: float = 0.5,
) -> (AsyncIterator[bytes], AsyncIterator[str]):
    """
    Data flow step that wraps a text to speech (TTS) step with buffering before the TTS and rate limiting on the output.

    Creates a data stream to handle a typical TTS flow.  It buffers input tokens to produce output as soon as possible but
    still keeping phrases together to produce natural sounding TTS.  It applies rate limiting to both the audio and text
    output.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator yielding text strings to be converted to speech.
    tts_step : TextToSpeechStep
        A function that takes an async iterator of text strings and returns an async iterator of audio bytes.
    buffer_seconds : float, default 0.5
        The duration (in seconds) of audio to buffer for rate limiting.

    Returns
    -------
    Tuple[AsyncIterator[bytes], AsyncIterator[str]]
        A tuple containing rate-limited iterators over the audio and text streams.

    Examples
    --------
    >>> tts_step = google_text_to_speech_step(
    ...     stream, text_to_speech_async_client, audio_format=AudioFormat.MP3
    ... )
    >>> stream = array_source(["Hello", "World"])
    >>> audio_stream, text_stream = tts_with_buffer_and_rate_limit_step(stream, tts_step)
    >>> audio_done = empty_sink(audio_stream)
    >>> text_done = array_sink(text_stream)
    >>> audio_out, text_out = asyncio.gather(audio_done, text_done)
    >>> assert text_out == [TimedText("Hello", 0.2), TimedText("World", 0.2)]
    """
    stream = buffer_tts_text_step(async_iter)
    stream = tts_step(stream)
    stream, audio_format_f = extract_value_step(stream, lambda x: x.audio_format)
    stream, text_output = tts_rate_limit_step(
        stream, audio_format_f, buffer_seconds=buffer_seconds
    )
    return stream, text_output


async def wait_for_punctuation_step(
    async_iter: AsyncIterator[str],
) -> AsyncIterator[str]:
    """
    Data flow step that buffers incoming text tokens into phrases based on punctuation.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator yielding text strings to be converted to speech.

    Returns
    -------
    AsyncIterator[str]
        The input token concatenated into phrases based on punctuation.
    """
    buffer = ""
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            buffer += item
            has_punc = any(char in string.punctuation for char in item.strip())
            if len(item) == 0 or has_punc:
                yield buffer
                buffer = ""


def tts_rate_limit_step(
    async_iter: AsyncIterator[AudioWithText],
    audio_format: AwaitableOrObj[AudioFormat],
    buffer_seconds: float = 0.5,
    include_text_output: bool = True,
) -> Union[AsyncIterator[bytes], Tuple[AsyncIterator[bytes], AsyncIterator[str]]]:
    """
    Applies rate limiting to an audio stream from a TTS output, optionally including synchronized text output.

    This function processes an asynchronous iterator of TTS audio output, each paired with text. It applies
    rate limiting to ensure that audio is emitted at a realistic rate based on its duration. Additionally,
    it can output timed text that is synchronized with the audio output. The text timing is adjusted based
    on the audio format and the buffer duration.  Rate-limiting allows the audio to be cancelled if an interruption
    occurs.

    Parameters
    ----------
    async_iter : AsyncIterator[voice_stream.AudioWithText]
        An asynchronous iterator yielding audio data paired with text.
    audio_format : AwaitableOrObj[voice_stream.audio.AudioFormat]
        The format of the audio stream, which can be an awaitable or a direct object.
    buffer_seconds : float, default 0.5
        The duration in seconds to buffer for rate limiting.
    include_text_output : bool, default True
        If True, includes a text output iterator that yields text synchronized with the audio rate.

    Returns
    -------
    Union[AsyncIterator[bytes], Tuple[AsyncIterator[bytes], AsyncIterator[str]]]
        Depending on `include_text_output`, either an async iterator of audio bytes or a tuple of
        async iterators for audio bytes and synchronized text.
    """

    async def tts_output_to_timed_text(tts: AudioWithText) -> TimedText:
        resolved_audio_format = await resolve_awaitable_or_obj(audio_format)
        duration = get_audio_length(resolved_audio_format, tts.audio)
        return TimedText(text=tts.text, duration_in_seconds=duration)

    if include_text_output:
        audio, text = fork_step(async_iter, pull_from_all=False)
        text = map_step(text, tts_output_to_timed_text)
        # text = log_step(text, "Timed text in")
        text = timed_text_rate_limit_step(text, buffer_seconds=buffer_seconds)
        # text = log_step(text, "Timed text out")
    else:
        audio = async_iter

    audio = map_step(audio, lambda x: x.audio)
    # audio = log_step(audio, "Audio out", lambda x: len(x))
    audio = audio_rate_limit_step(audio, audio_format, buffer_seconds=buffer_seconds)

    if include_text_output:
        return audio, text
    else:
        return audio


async def timed_text_rate_limit_step(
    async_iter: AsyncIterator[TimedText], buffer_seconds: float = 0.5
) -> AsyncIterator[str]:
    """
    Data flow step that takes :class:`~voice_stream.TimedText` objects and yield their text tokens based on the indicated timing.

    This step is usually used in combination with a rate limiting step.  The rate at which the text should be outputted
    is produced in a previous step and passed in `TimedText` objects.  This step then produces the raw text with the
    appropriate delays.

    Parameters
    ----------
    async_iter : AsyncIterator[TimedText]
        An asynchronous iterator over `TimedText` objects. Each `TimedText` object
        contains text and a duration in seconds.
    buffer_seconds : float, optional
        A buffer time in seconds to control the rate of text token output.  The buffer is designed so that the downstream
        iterator will always have at least buffer_seconds of text to display.

    Returns
    ------
    AsyncIterator[str]
        An asynchronous iterator yielding text tokens.

    Notes
    -----
    - The step will impose delays to slow down the rate at which objects pass through the stream, but because streams
      are pull-based, the downstream iterator must be ready to consume the objects that fast.  This step will never go
      faster than the downstream iterator allows.
    """
    last_send = time.perf_counter()
    queued_text_seconds = 0

    def compute_remaining(now):
        # Compute the amount of text remaining to be played in seconds.
        return max(0, queued_text_seconds - (now - last_send))

    async def break_into_tokens(timed_text: TimedText):
        tokens = timed_text.text.split(" ")
        if len(tokens) > 0:
            token_duration = timed_text.duration_in_seconds / len(tokens)
            tokens = [tokens[0]] + [" " + _ for _ in tokens[1:]]
            for token in tokens:
                yield TimedText(text=token, duration_in_seconds=token_duration)
        else:
            yield timed_text

    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for timed_text in owned_aiter:
            async for token in break_into_tokens(timed_text):
                now = time.perf_counter()
                remaining_text_seconds = compute_remaining(now)
                # If we have more than the buffer, sleep until we hit the buffer limit
                if remaining_text_seconds > buffer_seconds:
                    await asyncio.sleep(remaining_text_seconds - buffer_seconds)
                    # We don't know how long we actually slept.
                    now = time.perf_counter()
                    remaining_text_seconds = compute_remaining(now)
                queued_text_seconds = remaining_text_seconds + token.duration_in_seconds
                last_send = now
                yield token.text


async def min_tokens_step(
    async_iter: AsyncIterator[str], min_tokens: int
) -> AsyncIterator[str]:
    """
    Data flow step that aggregates string inputs to ensure a minimum number of tokens.

    The step assumes each input item is a single token, and concatenates the tokens until either an empty string is
    received or the number of concatenated tokens is at least `min_tokens`.

    Parameters
    ----------
    async_iter : AsyncIterator[TimedText]
        An asynchronous iterator over `TimedText` objects. Each `TimedText` object
        contains text and a duration in seconds.
    min_tokens : int
        The minimum number of tokens to buffer before outputting the concatenated value.

    Notes
    -----
    - If an empty token is detected, the function will automatically flush (i.e., clear and reset) the token list.
    """
    buffer = ""
    num_tokens = 0
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if len(item) == 0:
                if len(buffer) > 0:
                    yield buffer
                    buffer = ""
            else:
                buffer += item
                num_tokens += 1
                if num_tokens >= min_tokens:
                    yield buffer
                    buffer = ""
                    num_tokens = 0
    if len(buffer) > 0:
        yield buffer



import io
from enum import Enum

from mutagen.mp3 import MP3

from voice_stream.audio.audio_ogg import OggPage, OpusIdPacket


class AudioFormatError(Exception):
    """
    Indicates a problem related to audio formats.
    """

    pass


class AudioFormat(str, Enum):
    """
    A class representing an audio format as an Enum.
    """

    WAV_MULAW_8KHZ = "wav_mulaw_8khz"
    """WAV Telephone audio - mu-law encoded 8kHz."""

    OGG_OPUS = "ogg_opus"
    """Ogg container with Opus audio."""

    WEBM_OPUS = "webm_opus"
    """WebM container with Opus audio."""

    MP3 = "mp3"
    """MP3 audio format."""


def get_audio_length(audio_format: AudioFormat, audio: bytes):
    """
    Get the length of audio in seconds given the audio format and audio data.

    Parameters
    ----------
    audio_format : AudioFormat
        The format of audio data.
    audio : bytes
        The audio data.

    Returns
    -------
    float
        The length of the audio in seconds.

    Raises
    ------
    AudioFormatError
        If the audio format is not supported.
    """
    if audio_format == AudioFormat.WAV_MULAW_8KHZ:
        return len(audio) / 8000.0
    elif audio_format == AudioFormat.OGG_OPUS:
        first_page = OggPage.from_data(audio, 0)
        header = OpusIdPacket.from_ogg_page(first_page)
        page = OggPage.last_page_from_data(audio)
        return page.granule / float(header.input_sample_rate)
    elif audio_format == AudioFormat.MP3:
        return MP3(io.BytesIO(audio)).info.length
    else:
        raise AudioFormatError(
            f"Unsupported audio type '{audio_format}' for getting total audio length"
        )


def remove_wav_header(wav_bytes: bytes) -> bytes:
    """
    Removes the wav header from a wav file, regardless of the format.

    Parameters
    ----------
    wav_bytes
        The beginning of a WAV file, including the header.

    Returns
    -------
    bytes
        The audio bytes from the file, without the header.
    """
    # Check if it's a valid RIFF file
    if wav_bytes[:4] != b"RIFF" or wav_bytes[8:12] != b"WAVE":
        raise ValueError("Invalid WAV file")

    # Find the "data" chunk
    index = 12
    while index < len(wav_bytes):
        chunk_id = wav_bytes[index : index + 4]
        chunk_size = int.from_bytes(wav_bytes[index + 4 : index + 8], "little")
        if chunk_id == b"data":
            # Found the "data" chunk, return its content
            start = index + 8  # Skip "data" chunk header
            return wav_bytes[start : start + chunk_size]
        # logger.info(f"Skipping chunk {chunk_id}")
        index += 8 + chunk_size  # Move to the next chunk

    raise ValueError("WAV file does not contain a 'data' chunk")



from voice_stream.audio.audio_ops import (
    ogg_page_separator_step,
    ogg_concatenator_step,
    wav_mulaw_file_source,
    wav_mulaw_file_sink,
    audio_rate_limit_step,
)
from voice_stream.audio.audio_utils import (
    AudioFormat,
    AudioFormatError,
    get_audio_length,
    remove_wav_header,
)

__all__ = [
    "AudioFormat",
    "AudioFormatError",
    "audio_rate_limit_step",
    "get_audio_length",
    "ogg_page_separator_step",
    "ogg_concatenator_step",
    "remove_wav_header",
    "wav_mulaw_file_source",
    "wav_mulaw_file_sink",
]



from voice_stream.core import (
    array_sink,
    array_source,
    async_init_step,
    binary_file_sink,
    binary_file_source,
    concat_step,
    recover_exception_step,
    filter_step,
    fork_step,
    log_step,
    map_step,
    extract_value_step,
    merge_as_dict_step,
    merge_step,
    partition_step,
    queue_sink,
    queue_source,
    str_buffer_step,
    text_file_sink,
    text_file_source,
    count_step,
    byte_buffer_step,
    chunk_bytes_step,
    flatten_step,
    map_str_to_json_step,
    recover_exception_step,
    empty_source,
    none_source,
    QueueAsyncIterator,
    collect_dict_step,
    empty_sink,
)
from voice_stream.queue_with_exception import QueueWithException
from voice_stream.speech_to_text import (
    filter_spurious_speech_start_events_step,
    speech_with_start_detection_step,
)

from voice_stream.substreams import (
    substream_on_dict_key_step,
    substream_step,
    interruptable_substream_step,
    cancelable_substream_step,
)

from voice_stream.text_to_speech import (
    timed_text_rate_limit_step,
    buffer_tts_text_step,
    tts_rate_limit_step,
    tts_with_buffer_and_rate_limit_step,
    wait_for_punctuation_step,
    min_tokens_step,
    AudioWithText,
)
from voice_stream.audio.audio_ops import (
    audio_rate_limit_step,
    raw_audio_rate_limit_step,
)

__all__ = [
    "array_sink",
    "array_source",
    "async_init_step",
    "AudioWithText",
    "binary_file_sink",
    "binary_file_source",
    "buffer_tts_text_step",
    "byte_buffer_step",
    "cancelable_substream_step",
    "chunk_bytes_step",
    "collect_dict_step",
    "concat_step",
    "count_step",
    "empty_source",
    "empty_sink",
    "EndOfStreamMarker",
    "extract_value_step",
    "filter_step",
    "filter_spurious_speech_start_events_step",
    "flatten_step",
    "fork_step",
    "interruptable_substream_step",
    "log_step",
    "map_step",
    "map_str_to_json_step",
    "merge_as_dict_step",
    "merge_step",
    "min_tokens_step",
    "none_source",
    "partition_step",
    "queue_sink",
    "queue_source",
    "QueueAsyncIterator",
    "QueueWithException",
    "recover_exception_step",
    "speech_with_start_detection_step",
    "substream_on_dict_key_step",
    "substream_step",
    "text_file_sink",
    "text_file_source",
    "timed_text_rate_limit_step",
    "tts_rate_limit_step",
    "tts_with_buffer_and_rate_limit_step",
    "wait_for_punctuation_step",
]

from voice_stream.types import EndOfStreamMarker



import asyncio
import datetime
import logging
from typing import AsyncIterator, Callable, Tuple

import asyncstdlib

from voice_stream.core import log_step
from voice_stream.events import SpeechStart, BaseEvent, SpeechEnd, SpeechPartialResult

logger = logging.getLogger(__name__)

SpeechStep = Callable[
    [AsyncIterator[bytes]], Tuple[AsyncIterator[str], AsyncIterator[BaseEvent]]
]


def speech_with_start_detection_step(
    async_iter: AsyncIterator[bytes], speech_step: SpeechStep
):
    """
    Data flow step to perform speech recognition on a stream of audio and produce a robust start detection event.

    Takes a normal speech recognition step and filters the speech events to remove false :class:`~voice_stream.events.SpeechStart` events.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator that yields bytes of audio data.

    speech_step : SpeechStep
        A function that takes an async iterator as input and returns a tuple of a stream and speech events.

    Returns
    -------
    tuple
        A tuple of the audio stream and filtered speech start events.

    Example
    -------
    >>> speech_step = google_speech_v1_step(
    ...    stream,
    ...    speech_async_client,
    ...    audio_format=AudioFormat.WEBM_OPUS,
    ...    include_events=True,
    ... )
    >>> stream = fastapi_websocket_bytes_source(websocket)
    >>> stream, speech_events = speech_with_start_detection_step(stream, speech_step)
    >>> stream = merge_step(stream, speech_events)
    >>> await array_sink(stream)
    [SpeechStart(time_since_start=1.2), SpeechEnd(time_since_start=2.5), "Hello, how are you?"]
    """
    stream, speech_events = speech_step(async_iter)
    # speech_events = log_step(speech_events, "Speech event")
    speech_start = filter_spurious_speech_start_events_step(speech_events)
    return stream, speech_start


def filter_spurious_speech_start_events_step(
    async_iter: AsyncIterator[BaseEvent], threshold_secs: float = 1.0
):
    """
    Data flow step the filters a stream of speech events to remove false positives.

    This step filters speech events so that if a :class:`~voice_stream.events.SpeechStart` is quickly followed by a
    :class:`~voice_stream.events.SpeechEnd`, it is considered spurious and ignored.  This is useful for avoiding false
    detections which could create unnecessary interruptions.

    Parameters
    ----------
    async_iter : AsyncIterator[voice_stream.events.BaseEvent]
        An asynchronous iterator that yields bytes of audio data.

    threshold_secs : float
        The number of seconds to wait for a SpeechEnd after a SpeechStart is received.

    Returns
    -------
    AsyncIterator[voice_stream.events.BaseEvent]
        A modified event stream that removes the spurious :class:`~voice_stream.events.SpeechStart` events.

    Example
    -------
    >>> stream, speech_events = speech_step(async_iter)
    >>> speech_events = filter_spurious_speech_start_events_step(speech_events)

    Notes
    -------
    - Using this step does cause a delay in the movement of SpeechStart events down the stream.
    """

    class WaitingIterator:
        def __init__(self):
            self.aiter = async_iter.__aiter__()
            self.task_for_next_item = None

        def __aiter__(self):
            return self

        async def __anext__(self):
            end_time = None
            start_event = None
            while True:
                now = datetime.datetime.now()
                timeout_to_use = (
                    (end_time - now).microseconds / 1e6 if end_time else None
                )
                if timeout_to_use and timeout_to_use <= 0:
                    timeout_to_use = None
                if not self.task_for_next_item:
                    self.task_for_next_item = asyncio.create_task(
                        self.aiter.__anext__()
                    )
                # logger.debug(f"Starting to wait with timeout {timeout_to_use}.")
                done, pending = await asyncio.wait(
                    {self.task_for_next_item}, timeout=timeout_to_use
                )
                if pending:
                    # We didn't get an end event, so this was a real speech start.
                    # logger.debug("No response received in time.  Sending start event.")
                    assert not done
                    assert start_event
                    return start_event
                else:
                    assert done
                    self.task_for_next_item = None
                    item = await done.pop()
                    if isinstance(item, SpeechStart):
                        # logger.debug("Received Speech Start.  Resetting timer")
                        end_time = datetime.datetime.now() + datetime.timedelta(
                            seconds=threshold_secs
                        )
                        start_event = item
                    elif isinstance(item, SpeechEnd):
                        # logger.debug("Received Speech End, ignoring the start.")
                        end_time = None
                        start_event = None
                    else:
                        logger.info(f"Received unknown speech event, ignoring. {item}")

    stream = WaitingIterator()
    stream = log_step(stream, "New Speech Detected", lambda x: x.time_since_start)
    return stream


async def first_partial_speech_result_step(
    async_iter: AsyncIterator[BaseEvent],
) -> AsyncIterator[SpeechPartialResult]:
    """
    Data flow step that returns the first partial result after speech starts.

    This step uses partial results from the speech recognizer to signal the start of speech.  It waits for a
    :class:`~voice_stream.events.SpeechStart` event and then returns the first
    :class:`~voice_stream.events.SpeechPartialResult` that occurs after that.

    Parameters
    ----------
    async_iter : AsyncIterator[voice_stream.events.BaseEvent]
        An asynchronous iterator of Speech events

    Returns
    -------
    async_iter : AsyncIterator[voice_stream.events.SpeechPartialResult]
        The input stream filtered for just the first partial result after each speech start event.

    Example
    -------
    >>> async def gen():
    ...     yield SpeechStart(time_since_start=1)
    ...     yield SpeechEnd(time_since_start=2)
    ...     yield SpeechPartialResult(text="hello", time_since_start=1.5)
    ...     yield SpeechPartialResult(text="hello world", time_since_start=2)
    ...     yield SpeechStart(time_since_start=1)
    ...     yield SpeechPartialResult(text="next", time_since_start=2.5)
    >>> stream = gen()
    >>> stream = first_partial_speech_result_step(stream)
    >>> out = await array_sink(stream)
    >>> assert out == [
    ...     SpeechPartialResult(text="hello", time_since_start=1.5),
    ...     SpeechPartialResult(text="next", time_since_start=2.5)
    ... ]
    """
    started = False
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for item in owned_aiter:
            if started:
                if isinstance(item, SpeechPartialResult):
                    started = False
                    yield item
            elif isinstance(item, SpeechStart):
                started = True



"""
`Quart <https://github.com/pallets/quart/>`_ is an async Python web framework based on Flask.  VoiceStream provides
integrations to send and receive audio using websockets using Quart.

.. note::
    Flask generally uses SocketIO through the Flask_SocketIO library, which is different from regular WebSockets.
    If you just want regular websockets, it's easier to switch to Quart, which is a drop in replacement.

"""

import json
import logging
from asyncio import CancelledError
from typing import AsyncIterator, Union, Dict

try:
    from quart import websocket
except ImportError:
    raise ImportError(
        "Could not import quart python package. "
        "Please install it with `pip install quart`."
    )
from hypercorn.utils import UnexpectedMessageError


logger = logging.getLogger(__name__)


async def quart_websocket_source() -> AsyncIterator[Union[bytes, str]]:
    """
    Data flow source for receiving messages from a Quart WebSocket connection.

    This function facilitates the handling of incoming text or binary messages from a client connected through
    a Quart WebSocket. It continuously listens for messages and yields them as they arrive. The
    function handles the WebSocket connection lifecycle by accepting the connection and managing
    disconnection events.

    Returns
    -------
    AsyncIterator[Union[bytes, str]]
        An asynchronous iterator that yields incoming messages as bytes or strings.

    Notes
    -----
    - The function uses an infinite loop to listen for messages. It exits the loop and stops iterating
      when a CancelledError exception occurs.

    Examples
    --------
    >>> from quart import Quart
    >>> import asyncio
    >>>
    >>> app = Quart()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint():
    >>>     stream = quart_websocket_source(websocket)
    >>>     stream = map_step(stream, lambda x: "Echo: "+x)
    >>>     await quart_websocket_sink(stream)
    """
    await websocket.accept()
    while True:
        try:
            message = await websocket.receive()
        except CancelledError:
            logger.debug("websocket.receive is cancelled.")
            raise
        yield message


async def quart_websocket_sink(
    async_iter: AsyncIterator[Union[str, Dict, bytes]]
) -> None:
    """Takes an async iterator and sends everything to a Quart websocket.  Handles ."""
    """
    Data flow sink to strings, dicts, Pydantic objects, or bytes bytes to a Quart websocket.

    For each item in the incoming iterator, send it to the currently in scope Quart websocket connection.

    Parameters
    ----------
    async_iter : AsyncIterator[bytes]
        An asynchronous iterator that yields string, dictionaries, or bytes. Each incoming object is sent as a 
        websocket message.

    Returns
    -------
    None
        This function does not return any value.

    Examples
    --------
    >>> from quart import Quart
    >>> import asyncio
    >>>
    >>> app = Quart()
    >>>
    >>> @app.websocket("/ws")
    >>> async def websocket_endpoint():
    >>>     stream = quart_websocket_source(websocket)
    >>>     stream = map_step(stream, lambda x: "Echo: "+x)
    >>>     await quart_websocket_sink(stream)

    Notes
    -----
    - The function will continue sending data until the iterator is exhausted.
    - In case of any interruption or when the iterator is exhausted, it ensures proper cleanup.
    """
    try:
        async for message in async_iter:
            if isinstance(message, dict):
                message = json.dumps(message)
            elif hasattr(message, "model_dump"):
                message = message.model_dump()
            await websocket.send(message)
    except UnexpectedMessageError as e:
        if "ASGIWebsocketState.CLOSED" in str(e):
            logger.info("Websocket closed before all queued messages were sent.")
        else:
            raise e



import asyncio
import logging
from asyncio import CancelledError
from typing import (
    AsyncIterator,
    Callable,
    Tuple,
    Optional,
    List,
    Union,
    Any,
)

from voice_stream import QueueWithException
from voice_stream._substream_iters import SwitchableIterator, ResettableIterator
from voice_stream.core import (
    single_source,
    queue_source,
    concat_step,
    empty_sink,
    recover_exception_step,
    log_step,
    empty_source,
    queue_sink,
)
from voice_stream.types import (
    T,
    Output,
    to_source,
    to_tuple,
    from_tuple,
    SourceConvertable,
    EndOfStreamMarker,
    OptionalMultipleOutputs,
    format_current_task,
    cancel_with_confirmation,
    background_task,
)

logger = logging.getLogger(__name__)


async def substream_on_dict_key_step(
    async_iter: AsyncIterator[dict],
    key: str,
    substream_func: Callable[[AsyncIterator[Any]], AsyncIterator[Any]],
) -> AsyncIterator[dict]:
    """
    Data flow step that updates a value in a dictionary with the result of a substream.

    This step takes in a dictionary and produces a new dictionary that has one key modified.  The modified value comes
    from running a substream on the existing value.

    Parameters
    ----------
    async_iter : AsyncIterator[dict]
        An asynchronous iterator that yields dictionaries.
    key : str
        The key in the dictionary on which to perform the substreaming.
    substream_func : callable
        A function that takes an AsyncIterator and returns a stream based off that iterator.

    Returns
    -------
    AsyncIterator[dict]
        An asynchronous iterator that yields the modified dictionaries.

    Example
    -------
    >>> def substream(async_iter):
    ...     return map_step(async_iter, lambda x: x+2)
    >>> stream = array_source([{'a', 1, 'b':2}])
    >>> stream = substream_on_dict_key_step(stream, "b", substream)
    >>> out = await array_sink(stream)
    >>> assert out == [{'a', 1, 'b':4}]
    """
    current_dict = None

    async def input_gen():
        async for item in async_iter:
            nonlocal current_dict
            current_dict = item
            yield item[key]

    stream = input_gen()
    stream = substream_func(stream)
    async for item in stream:
        out = {**current_dict, key: item}
        yield out


async def substream_step(
    async_iter: AsyncIterator[T],
    substream_func: Callable[[AsyncIterator[T]], AsyncIterator[Output]],
) -> AsyncIterator[Output]:
    """
    Data flow step that runs a new stream on each item.

    A substream is useful when you want to group steps together for error handling or flow control.  This step calls the
    substream_func to create a new substream for each item from the source iterator.  The output of this step is the
    output of the substreams.  Each instance of the substream only gets one input value.

    Parameters
    ----------
    async_iter : AsyncIterator[T]
        An asynchronous iterator.
    substream_func : str
        A function that takes an AsyncIterator and creates a stream off of it.

    Returns
    -------
    AsyncIterator[T]
        An asynchronous iterator over the values produced by the substreams.

    Returns
    ----------
    >>> instance_count = 0
    >>> def substream(async_iter):
    ...     nonlocal instance_count
    ...     instance_count += 1
    ...     return map_step(async_iter, lambda x: x+instance_count)
    >>> stream = array_source([1,1,1])
    >>> stream = substream(stream, substream)
    >>> out = await array_sink(stream)
    >>> assert out == [2, 3, 4]
    """
    async for item in async_iter:
        stream = single_source(item)
        stream = substream_func(stream)
        async for sub_item in stream:
            yield sub_item


def cancelable_substream_step(
    async_iter: AsyncIterator[T],
    cancel_iter: AsyncIterator[T],
    substream_func: Callable[[AsyncIterator[T]], Union[AsyncIterator[Output], Tuple]],
    cancel_messages: Optional[List[SourceConvertable]] = None,
) -> OptionalMultipleOutputs:
    """
    Data flow step that runs a substream for each input, but takes a second iterator which causes the current substream to cancel.

    Calls the `substream_func` to create a new substream for each item from the source iterator.  If any item is produced
    from the `cancel_iter` during the processing of this substream, the substream is immediately stopped.  When a stop occurs,
    `cancel_messages` are optionally sent down each output stream.

    Parameters
    ----------
    async_iter
        The input AsyncIterator which we want to create substreams for.

    cancel_iter
        The cancel stream.  If an item appears on this iterator, the processing of the current substream is immediately stopped.

    substream_func
        The function used to generate AsyncIterators for each substream.

    cancel_messages
        An optional list of items to produce in the stream when a substream is cancelled.  If present, this must be a list
        which has the same length as the number of outputs returns by `substream_func`.  Each element will determine how cancels
        our signaled down that particular data stream.  If an AsyncIterator is passed, that iterator will be put into the stream.
        If any other object is passed, that signal object will be sent.  If None is passed, then nothing will be sent down the stream.

    Returns
    -------
     OptionalMultipleOutputs
        Either a single AsyncIterator or a tuple of multiple AsyncIterators.  The length is determined by the number of
        iterators returns from the substream_func.

    Notes
    -----
    - If you want to explicitly send `None` when a cancel occurs, use a :func:`~voice_stream:none_source`.

    """

    substream_completed = asyncio.Event()

    async def monitor_cancel():
        nonlocal next_substream, completed_outputs
        async for _ in cancel_iter:
            try:
                if completed_outputs < len(output_iters):
                    # logger.debug("Cancelling active substream.")
                    for i in output_iters:
                        await i.disconnect()
                    cancel_iters = _create_cancel_messages(
                        output_iters, cancel_messages
                    )
                    completed_outputs = 0
                    logger.debug("Substream canceled - sending cancel messages.")
                    await _switch_outputs(output_iters, cancel_iters)
                else:
                    logger.debug(
                        "Ignoring cancel because there is no active substream."
                    )
            except Exception as e:
                logger.error(f"Error in cancelable_substream: {e}", exc_info=True)

    def on_output_complete():
        nonlocal completed_outputs
        completed_outputs += 1
        if completed_outputs == len(output_iters):
            # logger.debug("Marking substream completed")
            substream_completed.set()

    next_source = queue_source()
    next_substream = to_tuple(substream_func(next_source))
    output_iters = [
        SwitchableIterator(
            None, callback=on_output_complete, propagate_end_of_iter=False
        )
        for _ in next_substream
    ]
    completed_outputs = len(output_iters)
    cancel_messages = (
        cancel_messages if cancel_messages else [None for _ in output_iters]
    )

    monitor_cancel_task = background_task(monitor_cancel())

    async def gen():
        nonlocal next_source, next_substream, completed_outputs
        async for item in async_iter:
            # logger.debug("Starting substream")
            completed_outputs = 0
            substream_completed.clear()
            await _switch_outputs(output_iters, next_substream)
            await next_source.put(item)
            await next_source.put(EndOfStreamMarker)
            await substream_completed.wait()
            # logger.debug("Completed substream")
            next_source = queue_source()
            next_substream = to_tuple(substream_func(next_source))
        # logger.debug("Exhausted input iterator")
        monitor_cancel_task.cancel()
        await next_source.put(EndOfStreamMarker)
        await asyncio.wait([asyncio.create_task(empty_sink(i)) for i in next_substream])
        for i in output_iters:
            await i.end_iteration()

    background_task(gen())
    return from_tuple(output_iters)


def interruptable_substream_step(
    async_iter: AsyncIterator[T],
    substream_func: Callable[
        [AsyncIterator[T]],
        Callable[[AsyncIterator[T]], Union[AsyncIterator[Output], Tuple]],
    ],
    cancel_messages: Optional[List[SourceConvertable]] = None,
) -> OptionalMultipleOutputs:
    """
    Data flow step that creates a substream which will get interrupted if a new value comes in.

    For each input, creates a new substream and runs it.  If a new input is available before the substream completes,
    this cancels the existing substream and starts a new one.  This is similar to :func:`~voice_stream.cancelable_substream_step`
    except that it uses the same iterator for input values and to trigger cancellation.

    Parameters
    ----------
    async_iter
        The input AsyncIterator which we want to create substreams for.

    substream_func
        The function used to generate AsyncIterators for each substream.

    cancel_messages
        An optional list of items to produce in the stream when a substream is cancelled.  If present, this must be a list
        which has the same length as the number of outputs returns by `substream_func`.  Each element will determine how cancels
        our signaled down that particular data stream.  If an AsyncIterator is passed, that iterator will be put into the stream.
        If any other object is passed, that signal object will be sent.  If None is passed, then nothing will be sent down the stream.

    Returns
    -------
     OptionalMultipleOutputs
        Either a single AsyncIterator or a tuple of multiple AsyncIterators.  The length is determined by the number of
        iterators returns from the substream_func.
    """

    completed_outputs = 0

    def on_output_complete():
        nonlocal completed_outputs
        completed_outputs += 1

    active_source = None
    next_source = queue_source()
    next_substream = to_tuple(substream_func(next_source))
    # logger.debug(f"Created substream {next_substream}")
    output_iters = [
        SwitchableIterator(_, callback=on_output_complete, propagate_end_of_iter=False)
        for _ in next_substream
    ]

    async def gen():
        nonlocal active_source, next_source, next_substream, completed_outputs
        async for item in async_iter:
            if active_source:
                if cancel_messages and completed_outputs < len(output_iters):
                    cancel_iters = _create_cancel_messages(
                        output_iters, cancel_messages
                    )
                    next_substream = [
                        concat_step(cancel_iter, output)
                        for output, cancel_iter in zip(next_substream, cancel_iters)
                    ]
                completed_outputs = 0
                await _switch_outputs(output_iters, next_substream)
            # logger.debug("Running substream")
            active_source = next_source
            await active_source.put(item)
            await active_source.put(EndOfStreamMarker)
            next_source = queue_source()
            next_substream = to_tuple(substream_func(next_source))
        # logger.debug("Completed interruptable substream iter")
        for i in output_iters:
            await i.end_iteration()

    # logger.debug("Creating task")
    background_task(gen())
    return from_tuple(output_iters)


def exception_handling_substream(
    async_iter: AsyncIterator[T],
    substream_func: Callable[
        [AsyncIterator[T]],
        Callable[[AsyncIterator[T]], Union[AsyncIterator[Output], Tuple]],
    ],
    exception_handler: Callable[[Exception], List[Any]],
    num_outputs: int = 1,
    max_exceptions: Optional[int] = None,
):
    input_queue = QueueWithException(maxsize=1)
    end_of_input = False

    async def mark_end_of_iteration(async_iter: AsyncIterator):
        try:
            async for item in async_iter:
                yield item
        finally:
            nonlocal end_of_input
            end_of_input = True

    input_stream = mark_end_of_iteration(async_iter)
    input_task = background_task(
        input_queue.enqueue_iterator(input_stream, include_end_of_stream=True)
    )
    queues = [QueueWithException(maxsize=1) for _ in range(num_outputs)]
    outputs = [queue_source(q) for q in queues]

    async def gen():
        try:
            exception_count = 0
            last_exception = None
            while (not max_exceptions) or exception_count <= max_exceptions:
                if end_of_input and input_queue.qsize() == 0:
                    for queue in queues:
                        await queue.put(EndOfStreamMarker)
                    return
                else:
                    substream = queue_source(input_queue)
                    substream_outputs = to_tuple(substream_func(substream))
                    if len(outputs) != len(substream_outputs):
                        raise ValueError(
                            f"num_outputs was {num_outputs}, but substream_func returned {len(substream_outputs)} outputs."
                        )
                    tasks = [
                        background_task(
                            q.enqueue_iterator(
                                s,
                                include_end_of_stream=False,
                                propagate_exception=False,
                            )
                        )
                        for q, s in zip(queues, substream_outputs)
                    ]
                    try:
                        await asyncio.gather(*tasks)
                    except Exception as e:
                        last_exception = e
                        logger.exception("Exception in substream")
                        exception_count += 1
                        cancel_iters = _create_cancel_messages(
                            queues, exception_handler(e)
                        )
                        # logger.debug("Enqueuing exception tasks")
                        exception_tasks = [
                            background_task(
                                q.enqueue_iterator(
                                    s,
                                    include_end_of_stream=False,
                                    propagate_exception=True,
                                )
                            )
                            for q, s in zip(queues, cancel_iters)
                        ]
                        await asyncio.gather(*exception_tasks)
                        await asyncio.sleep(1.0)
                        # logger.debug("Restarting substream after exception.")
            for q in queues:
                await q.set_exception(last_exception)
            # logger.debug(f"Exceeded the limit of {max_exceptions} exception recoveries.  Got {exception_count}. Terminating")
        finally:
            # logger.info("Exiting gen()")
            await cancel_with_confirmation(input_task)

    background_task(gen())
    return from_tuple(outputs)


def _create_cancel_messages(
    outputs: List[AsyncIterator], cancel_messages: Optional[List[SourceConvertable]]
):
    """
    The cancel_messages list must have the same length as the outputs list, indicating which output to cancel. If cancel_messages is not provided, it defaults to None.

    The output cancel messages are converted to the appropriate type using the to_source() function.
    """
    if cancel_messages and len(cancel_messages) != len(outputs):
        raise ValueError(
            f"cancel_messages must be the same length as the number of outputs for the substream.  Got {len(cancel_messages)} cancel messages and {len(outputs)} substream outputs."
        )

    return [to_source(_) for _ in cancel_messages]


async def _switch_outputs(
    switchable_iters: List[SwitchableIterator], new_outputs: List[AsyncIterator]
):
    """
    Switch the outputs of `switchable_iters` with the provided `new_outputs`.
    """
    assert len(new_outputs) == len(switchable_iters)
    for switchable_iter, new_output in zip(switchable_iters, new_outputs):
        await switchable_iter.switch(new_output)



"""
Additional utilities for working with the Google Cloud steps.
"""

import logging
from typing import Union, Optional

from google.cloud.speech_v2 import (
    Recognizer,
    CreateRecognizerRequest,
    SpeechAsyncClient,
    RecognitionConfig,
    ExplicitDecodingConfig,
    AutoDetectDecodingConfig,
)

from voice_stream.audio.audio_ops import AudioFormat

GoogleDecodingConfig = Optional[Union[AudioFormat, ExplicitDecodingConfig]]

logger = logging.getLogger(__name__)


async def create_recognizer(
    client: SpeechAsyncClient,
    project_id: str,
    location: str,
    recognizer_id: str,
    model: str,
    language_codes: Union[str, list[str]],
    audio_format: GoogleDecodingConfig = None,
):
    """
    Helper function to create a Google Speech V2 recognizer in a Google Cloud project.

    With Google Speech V2, you need to create a recognizer first before you can run speech recognition.  The V1 API
    uses a fixed set of recognizers and doesn't require this step.  This function sets up and initializes a speech
    recognizer with the specified configuration options.

    Parameters
    ----------
    client : SpeechAsyncClient
        An instance of SpeechAsyncClient for interacting with the Google Cloud Speech API.
    project_id : str
        The ID of the Google Cloud project where the recognizer is to be created.
    location : str
        The location or region where the recognizer is to be deployed.
    recognizer_id : str
        The unique identifier for the recognizer within the project.
    model : str
        The model to be used by the recognizer for speech recognition.
    language_codes : Union[str, list[str]]
        The language code(s) for the recognizer. Can be a single string or a list of strings.
    audio_format : GoogleDecodingConfig, optional
        Optional configuration for audio decoding. If not specified, the recognizer will auto-detect the audio configuration.

    Notes
    -----
    - The function handles the creation and configuration of a recognizer in an
      asynchronous manner, utilizing the `SpeechAsyncClient`.
    - It constructs the necessary request with the provided parameters and sends it
      to the Google Cloud Speech API.
    - After the recognizer is successfully created, the function logs its name for
      confirmation.

    Examples
    --------
    >>> await create_recognizer(client, "my-project", "us-west1", "my-recognizer", "phone_call", "en-US")

    This example demonstrates how to create a recognizer for English language
    phone calls in the 'us-west1' region of the 'my-project' Google Cloud project.
    """
    # Create a recognizer configuration
    parent = f"projects/{project_id}/locations/{location}"
    auto_decoding, explicit_decoding = _resolve_audio_decoding(audio_format)

    # Create the recognizer
    # noinspection PyTypeChecker
    operation = client.create_recognizer(
        request=CreateRecognizerRequest(
            parent=parent,
            recognizer=Recognizer(
                name=f"{parent}/recognizers/{recognizer_id}",
                default_recognition_config=RecognitionConfig(
                    auto_decoding_config=auto_decoding,
                    explicit_decoding_config=explicit_decoding,
                    model=model,
                    language_codes=language_codes,
                ),
            ),
            recognizer_id=recognizer_id,
        )
    )

    # Wait for the operation to complete
    response = (await operation).result(timeout=90)
    response = await response
    logger.info(f"Recognizer created: {response.name}")


def _resolve_audio_decoding(audio_format):
    if isinstance(audio_format, ExplicitDecodingConfig):
        return None, audio_format
    elif audio_format == AudioFormat.WAV_MULAW_8KHZ:
        # noinspection PyTypeChecker
        explicit = ExplicitDecodingConfig(
            encoding=ExplicitDecodingConfig.AudioEncoding.MULAW,
            sample_rate_hertz=8000,
            audio_channel_count=1,
        )
        return None, explicit
    else:
        return AutoDetectDecodingConfig(), None



import asyncio
import logging
import traceback
from asyncio import InvalidStateError
from typing import AsyncIterator, TypeVar, Optional

from voice_stream import queue_source, QueueWithException
from voice_stream.types import cancel_with_confirmation, EndOfStreamMarker, format_task

T = TypeVar("T")

logger = logging.getLogger(__name__)


class ResettableIterator(AsyncIterator):
    """Wraps an AsyncIterator and allows it to be used in multiple downstream generators.  It works like a normal
    iterator, but if 'reset' is called, it throws a StopAsyncIteration.  This ends iteration for all downstream tasks.
     However, the underlying iterator isn't finished and so this iterator can be used in another generator.
    """

    def __init__(self, async_iter: AsyncIterator):
        self.async_iter = async_iter
        self.reset_flag = asyncio.Event()
        self.next_item_task = None
        self.lock = asyncio.Lock()

    def __aiter__(self):
        return self

    async def __anext__(self):
        async with self.lock:
            # logger.trace(f"Entering __anext__")
            assert (
                not self.reset_flag.is_set()
            ), f"Reset flag should be set on entry.  ResettableIterator isn't multithreaded"
            if self.next_item_task is None:
                self.next_item_task = asyncio.create_task(self.async_iter.__anext__())
            reset_task = asyncio.create_task(self.reset_flag.wait())
            done, pending = await asyncio.wait(
                {self.next_item_task, reset_task}, return_when=asyncio.FIRST_COMPLETED
            )
            # logger.trace(f"Tasks completed: Done is {done}.  Pending is {pending}")
            if reset_task in done:
                logger.debug(f"Reset flag set on wait")
                raise StopAsyncIteration
            assert reset_task in pending and self.next_item_task in done
            await cancel_with_confirmation(reset_task)
            out = await self.next_item_task
            self.next_item_task = None
            return out

    def reset(self):
        self.reset_flag.set()
        # It's important that we create a new event because we don't know if __anext__ will be called again before the original event is cleared.
        self.reset_flag = asyncio.Event()

    async def confirm_reset(self):
        async with self.lock:
            return True


class SwitchableIterator:
    """
    An iterator that can change its source.

    It is created with an iterator.  When disconnect is called, it cancels looking at the original iterator and waits for a call to switch.
    When switch is called, moves over to the pull from the new iterator.

    If None is passed for the initial async_iter, then nothing will happen until switch is called.

    The callback will be called if the current iterator completes.  Not called on a cancellation.

    The state model for the initial iterator

    Connected - The iterator is actively forwarding from a source iterator.  Any call to __anext__ will be forwarded to that iterator.
    Disconnected - The is not connected.  Any call to __anext__ will block until the iterator is connected.
    Completed - The iterator is done, it has sent an end of iteration message.
    """

    CONNECTED = "Connected"
    DISCONNECTED = "Disconnected"
    COMPLETED = "Completed"

    def __init__(
        self,
        async_iter: Optional[AsyncIterator],
        callback=None,
        propagate_end_of_iter: bool = True,
    ):
        self.callback = callback
        self.propagate_end_of_iter = propagate_end_of_iter
        self.async_iter = None
        self.state = SwitchableIterator.DISCONNECTED
        self.task = None
        self.queue = QueueWithException(maxsize=1)
        self.queue_source = queue_source(self.queue)
        if async_iter:
            self._new_input(async_iter)

    def __aiter__(self):
        return self.queue_source

    async def end_iteration(self):
        if self.state == SwitchableIterator.COMPLETED:
            raise InvalidStateError("Cannot switch a completed iterator")
        await self.disconnect()
        self.state = SwitchableIterator.COMPLETED
        self.async_iter = None
        self.task = asyncio.create_task(self.queue.put(EndOfStreamMarker))

    async def switch(self, async_iter: AsyncIterator):
        if self.state == SwitchableIterator.COMPLETED:
            raise InvalidStateError("Cannot switch a completed iterator")
        await self.disconnect()
        self._new_input(async_iter)

    def _callback(self, task):
        assert task == self.task
        self.task = None
        self.async_iter = None
        logger.debug(
            f"Switchable iterator - completed listener task {format_task(task)}"
        )
        if self.callback and not task.cancelled():
            assert task.done()
            self.callback()

    def _new_input(self, async_iter: AsyncIterator):
        self.async_iter = async_iter
        coro = self.queue.enqueue_iterator(
            async_iter, include_end_of_stream=self.propagate_end_of_iter
        )
        task = asyncio.create_task(coro)
        logger.debug(f"Switchable iterator - new listener task {format_task(task)}")
        self.task = task
        self.task.add_done_callback(self._callback)

    async def disconnect(self):
        if self.async_iter:
            assert self.task
            logger.debug(
                f"Switchable iterator - cancelling listener {format_task(self.task)}"
            )

            await cancel_with_confirmation(self.task)
        self.async_iter = None
        self.state = SwitchableIterator.DISCONNECTED


# class SwitchableIteratorOld:
#     """
#     An iterator that can change its source.
#
#     It is created with an iterator.  When disconnect is called, it cancels looking at the original iterator and waits for a call to switch.
#     When switch is called, moves over to the pull from the new iterator.
#
#     If None is passed for the initial async_iter, then nothing will happen until switch is called.
#
#     The callback allows you to specify whether to propagate a StopAsyncIteration exception.  If you don't propagate, then you must eventually call end_iteration.
#
#     The state model for the initial iterator
#
#     Connected - The iterator is actively forwarding from a source iterator.  Any call to __anext__ will be forwarded to that iterator.
#     Disconnected - The is not connected.  Any call to __anext__ will block until the iterator is connected.
#     Completed - The iterator is done, it has sent an end of iteration message.
#     """
#
#     CONNECTED = "Connected"
#     DISCONNECTED = "Disconnected"
#     COMPLETED = "Completed"
#
#     def __init__(
#             self,
#             async_iter: Optional[AsyncIterator],
#             callback=None,
#             propagate_end_of_iter: bool = True,
#     ):
#         self.state = (
#             SwitchableIterator.CONNECTED
#             if async_iter
#             else SwitchableIterator.DISCONNECTED
#         )
#         self.state_change_event = asyncio.Event()
#         self.state_change_completed_event = asyncio.Event()
#         self.async_iter = async_iter
#         self.callback = callback
#         self.propagate_end_of_iter = propagate_end_of_iter
#         # logger.debug(f"New SwitchableIterator {self}")
#
#     def __aiter__(self):
#         return self
#
#     async def __anext__(self):
#         # logger.debug(f"Entering SwitchableIterator.__anext__ {self}")
#         log_next_iter = False
#         while True:
#             # logger.debug(f"Looping with state {self.state}")
#             self.state_change_event.clear()
#             self.state_change_completed_event.set()
#             try:
#                 if self.state == SwitchableIterator.DISCONNECTED:
#                     await self.state_change_event.wait()
#                 elif self.state == SwitchableIterator.COMPLETED:
#                     raise StopAsyncIteration
#                 else:
#                     assert self.state == SwitchableIterator.CONNECTED
#                     next_item_task = asyncio.create_task(self.async_iter.__anext__())
#                     state_change_task = asyncio.create_task(
#                         self.state_change_event.wait()
#                     )
#                     if log_next_iter:
#                         logger.debug("Starting to wait for next item")
#                     done, pending = await asyncio.wait(
#                         {next_item_task, state_change_task},
#                         return_when=asyncio.FIRST_COMPLETED,
#                     )
#                     if log_next_iter:
#                         log_next_iter = False
#                         logger.debug(
#                             f"Switchable iterator - Done: {done} Pending: {pending}"
#                         )
#                     for task in pending:
#                         task.cancel()
#                         # try:
#                         #     await task
#                         # except asyncio.CancelledError:
#                         #     if task == next_item_task:
#                         #         logger.debug(f"Cancelled next item task {task}")
#                         #     raise
#                         # except StopAsyncIteration:
#                         #     pass
#                     if next_item_task in done:
#                         try:
#                             return await next_item_task
#                         except StopAsyncIteration as e:
#                             # logger.debug(f"Substream completed")
#                             if self.callback:
#                                 self.callback()
#                             if self.propagate_end_of_iter:
#                                 raise e
#                             # If we aren't propagating the end, disconnect the iterator.
#                             self.disconnect()
#                     else:
#                         logger.debug(
#                             f"Iterators switched during __anext__().  Setting {self.state_change_completed_event}"
#                         )
#                         log_next_iter = True
#                         pass
#             finally:
#                 self.state_change_event.clear()
#                 do_log = not self.state_change_completed_event.is_set()
#                 self.state_change_completed_event.set()
#                 if do_log:
#                     logger.debug(f"Setting {self.state_change_completed_event} done")
#
#     def end_iteration(self):
#         self.state = SwitchableIterator.COMPLETED
#         self.async_iter = None
#         self.state_change_completed_event.clear()
#         self.state_change_event.set()
#
#     def switch(self, async_iter: AsyncIterator):
#         if not inspect.isasyncgen(async_iter) and not is_async_iterator(async_iter):
#             raise ValueError(
#                 f"SwitchableIterator.switch must be called with an AsyncIterator.  Got {async_iter}"
#             )
#         # logger.debug("Switching iterators")
#         if self.state == SwitchableIterator.COMPLETED:
#             raise InvalidStateError("Cannot switch a completed iterator")
#         self.state = SwitchableIterator.CONNECTED
#         self.async_iter = async_iter
#         self.state_change_completed_event.clear()
#         self.state_change_event.set()
#
#     def disconnect(self):
#         # logger.debug("Disconnecting iterators")
#         if self.state == SwitchableIterator.COMPLETED:
#             raise InvalidStateError("Cannot disconnect a completed iterator")
#         self.state = SwitchableIterator.DISCONNECTED
#         self.async_iter = None
#         self.state_change_completed_event.clear()
#         self.state_change_event.set()
#
#     async def wait_for_state_change(self):
#         logger.debug(
#             f"Waiting for state change completion {self.state_change_completed_event}"
#         )
#         await self.state_change_completed_event.wait()
#
#
# class SimpleFlow(AsyncIterator):
#     """Serves as a stub for creating flows.  Acts like an async iterator, but when apply is called with an async iterator, it wraps that iterator."""
#
#     def __init__(self):
#         loop = asyncio.get_running_loop()
#         self.iter_f = loop.create_future()
#         self.iter = None
#
#     def finish(self, *outputs: AsyncIterator[Any]):
#         self.outputs = outputs
#
#     def __call__(self, async_iter: AsyncIterator[T]):
#         if not self.outputs:
#             raise InvalidStateError("Flow was not finished with finish()")
#         self.iter_f.set_result(async_iter)
#         return self.outputs
#
#     def flow_for_output(self, ix: int):
#         def output_wrapper(async_iter: AsyncIterator[T]):
#             out = self.__call__(async_iter)
#             return out[ix]
#
#         return output_wrapper
#
#     def __aiter__(self):
#         if self.iter:
#             self.iter = self.iter_f.result().__aiter__()
#         return self
#
#     async def __anext__(self):
#         if not self.iter:
#             logger.info(
#                 "Stub iterator not resolved before use.  If the stream stalls, this could be the problem."
#             )
#             self.iter = (await self.iter_f).__aiter__()
#         return await self.iter.__anext__()



import logging
from pprint import pformat
from typing import Dict, Iterator

from langchain_core.runnables import RunnableBranch
from langchain_core.runnables.base import RunnableLike, RunnableGenerator

from voice_stream.types import T

logger = logging.getLogger(__name__)


class RunnableSwitch(RunnableBranch):
    """Like RunnableBranch, but just takes a dict"""

    def __init__(
        self,
        input_key: str,
        branches: Dict[
            str,
            RunnableLike,
        ],
        default_branch: RunnableLike,
    ) -> None:
        def check(y):
            def f(x):
                # logger.debug(f"Checking {x[input_key]} == {y}")
                return x[input_key] == y

            return f

        b = [*[(check(k), v) for k, v in branches.items()], default_branch]
        super().__init__(*b)


def RunnableLogger(header: str):
    """
    Logs out whatever is coming through the pipeline.

    Parameters
    ----------
    header
        Message to prepend to the log.

    Yields
    ------
    Yields the same items that come in.
    """

    async def async_gen(input: Iterator[T]) -> Iterator[T]:
        # hold partial input until we get a comma
        async for entry in input:
            logger.info(f"{header} {type(entry)}\n{pformat(entry)}")
            yield entry

    def gen(input: Iterator[T]) -> Iterator[T]:
        # hold partial input until we get a comma
        for entry in input:
            logger.info(f"{header} {type(entry)}\n{pformat(entry)}")
            yield entry

    return RunnableGenerator(gen, atransform=async_gen)
